"""
Image grid saver, based on color_grid_vis from github.com/Newmu
"""

import numpy as np
import scipy.misc
from scipy.misc import imsave

def save_images(X, save_path):
    # [0, 1] -> [0,255]
    if isinstance(X.flatten()[0], np.floating):
        X = (255.99*X).astype('uint8')

    n_samples = X.shape[0]
    rows = int(np.sqrt(n_samples))
    while n_samples % rows != 0:
        rows -= 1

    nh, nw = rows, n_samples/rows

    if X.ndim == 2:
        X = np.reshape(X, (X.shape[0], int(np.sqrt(X.shape[1])), int(np.sqrt(X.shape[1]))))

    if X.ndim == 4:
        # BCHW -> BHWC
        X = X.transpose(0,2,3,1)
        h, w = X[0].shape[:2]
        img = np.zeros((int(h*nh), int(w*nw), 3))
    elif X.ndim == 3:
        h, w = X[0].shape[:2]
        img = np.zeros((int(h*nh), int(w*nw)))

    for n, x in enumerate(X):
        j = int(n/nw)
        i = int(n%nw)
        img[j*h:j*h+h, i*w:i*w+w] = x

    imsave(save_path, img)



import sys
import time
import jsonschema

from pyneval.metric.utils.config_utils import DINF
from pyneval.model.swc_node import SwcTree
from pyneval.metric.utils.km_utils import KM, get_dis_graph
from pyneval.metric.utils.point_match_utils import get_swc2swc_dicts

from pyneval.metric.utils.metric_manager import get_metric_manager

metric_manager = get_metric_manager()

class LinkMetric(object):
    """
    link metric
    """

    def __init__(self, config):
        self.scale = config['scale']
        self.debug = config["debug"] if config.get("debug") is not None else False

    def get_neighbors(self, node):
        neighbors = list(node.children)
        if node.parent is not None and not node.parent.is_virtual():
            neighbors.append(node.parent)
        return neighbors


    def get_same_pos_nodes(self, gold_nodes, test_nodes):
        res = []
        for gold_node in gold_nodes:
            for test_node in test_nodes:
                if gold_node.get_center_as_tuple() == test_node.get_center_as_tuple():
                    res.append(gold_node)
                    break
        return res


    def get_extra_pos_nodes(self, big_set, subset):
        res = []
        pos_set = set()
        for node in subset:
            pos_set.add(node.get_center_as_tuple())

        for node in big_set:
            if not node.get_center_as_tuple() in pos_set:
                res.append(node)
        return res

    def run(self, gold_swc_tree, test_swc_tree):
        gold_swc_tree.rescale(self.scale)
        test_swc_tree.rescale(self.scale)

        gold_list = gold_swc_tree.get_node_list()
        test_list = test_swc_tree.get_node_list()

        gold_test_dict = get_swc2swc_dicts(src_node_list=gold_swc_tree.get_node_list(),
                                           tar_node_list=test_swc_tree.get_node_list())
        test_gold_dict = get_swc2swc_dicts(src_node_list=test_swc_tree.get_node_list(),
                                           tar_node_list=gold_swc_tree.get_node_list())

        edge_loss, tree_dis_loss = 0.0, 0.0
        for gold_node in gold_list:
            if gold_node.is_virtual():
                continue

            test_node = gold_test_dict[gold_node]

            # attention: nodes in test neighbors are labeled by TEST tree label
            test_neighbors = self.get_neighbors(test_node)
            gold_neighbors = self.get_neighbors(gold_node)

            nodes_in_same_pos = self.get_same_pos_nodes(gold_neighbors, test_neighbors)
            nodes_gold_extra = self.get_extra_pos_nodes(gold_neighbors, nodes_in_same_pos)
            # attention: nodes in test neighbors are labeled by TEST tree label
            nodes_test_extra = self.get_extra_pos_nodes(test_neighbors, nodes_in_same_pos)

            for gold_extra in nodes_gold_extra:
                edge_loss += gold_extra.distance(gold_node)

            for test_extra in nodes_test_extra:
                map_test_extra = test_gold_dict[test_extra]
                edge_loss += map_test_extra.distance(gold_node)

            # if two nodes are not connected, the dis will be set as DINF, so the threshold need to be set as DINF - 1
            dis_graph, switch, test_len, gold_len = \
                get_dis_graph(gold_tree=gold_swc_tree, test_tree=test_swc_tree,
                              gold_node_list=nodes_gold_extra, test_node_list=nodes_test_extra,
                              test_gold_dict=test_gold_dict, threshold_dis=DINF - 1, metric_mode=2)

            km = KM(maxn=max(test_len, gold_len) + 10, nx=test_len, ny=gold_len, G=dis_graph)
            km.solve()
            tree_dis_loss += -km.get_max_dis()

        res = {
            "edge_loss": edge_loss,
            "tree_dis_loss": tree_dis_loss
        }
        return res, None, None

@metric_manager.register(
    name="link",
    config="link_metric.json",
    desc="",
    public=False,
    alias=['LM'],
)
def link_metric(gold_swc_tree, test_swc_tree, config):
    link = LinkMetric(config)
    return link.run(gold_swc_tree, test_swc_tree)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2019/8/17
# @Author  : github.com/guofei9987

# @Time    : 2021/8/29
# @Author  : zhanghan
import copy
import os
import numpy as np
import multiprocessing as mp


def get_default_x0(parameters):
    res = []
    for para in parameters:
        res.append(parameters[para]["default"])
    return res

def set_default_x0(parameters, x_new_default):
    it = 0
    for para in parameters:
        parameters[para]["default"] = x_new_default[it]
        it += 1

class SimulatedAnnealingBase():
    """
    DO SA(Simulated Annealing)

    Parameters
    ----------------
    func : function
        The func you want to do optimal
    n_dim : int
        number of variables of func
    x0 : array, shape is n_dim
        initial solution
    T_max :float
        initial temperature
    T_min : float
        end temperature
    L : int
        num of iteration under every temperature（Long of Chain）

    Attributes
    ----------------------


    Examples
    -------------
    See https://github.com/guofei9987/scikit-opt/blob/master/examples/demo_sa.py
    """
    def __init__(self, func, gold_swc_tree, metric_method, 
                 metric_config, optimize_config,**kwargs):
        assert optimize_config["optimize"]["SA"]["Tmax"] > optimize_config["optimize"]["SA"]["Tmin"] > 0, 'Tmax > Tmin > 0'


        self.func = func
        self.T_max = optimize_config["optimize"]["SA"]["Tmax"]  # initial temperature
        self.T_min = optimize_config["optimize"]["SA"]["Tmin"]  # end temperature
        self.L = optimize_config["optimize"]["SA"]["L"]  # num of iteration under every temperature（also called Long of Chain）
        # stop if best_y stay unchanged over max_stay_counter times (also called cooldown time)
        self.max_stay_counter = optimize_config["optimize"]["SA"]["maxStayCounter"]
        self.q = optimize_config["optimize"]["SA"]["q"]
        parameters = optimize_config["trace"]["parameters"]
        self.n_dims = len(parameters)
        
        self.best_x = parameters  # initial solution
        self.best_y = self.func(gold_tree=gold_swc_tree, metric_method=metric_method, config=get_default_x0(parameters),
                                metric_config=metric_config, optimize_config=optimize_config,
                                test_name="test_init", lock=None)[1]
        self.T = self.T_max
        self.iter_cycle = 0
        self.generation_best_X, self.generation_best_Y = [self.best_x], [self.best_y]
        # history reasons, will be deprecated
        self.best_x_history, self.best_y_history = self.generation_best_X, self.generation_best_Y

    def get_new_x(self, x):
        u = np.random.uniform(-1, 1, size=self.n_dims)
        x_new = x + 20 * np.sign(u) * self.T * ((1 + 1.0 / self.T) ** np.abs(u) - 1.0)
        return x_new

    def cool_down(self):
        self.T = self.T * self.q

    def isclose(self, a, b, rel_tol=1e-09, abs_tol=1e-30):
        return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)

    def run(self, gold_swc_tree, metric_method, metric_config, optimize_config):
        x_current, y_current = self.best_x, self.best_y
        stay_counter = 0
        CPU_CORE_NUM = 15
        while True:
            # loop L times under the same Temperature
            i = 0
            print("[Info: ]x current = {}".format(get_default_x0(x_current)))

            while i < int(self.L):
                pool = mp.Pool(processes=CPU_CORE_NUM)
                res_x, res_y = [], []
                lock = mp.Manager().Lock()
                for j in range(CPU_CORE_NUM):
                    x_new = self.get_new_x(x_current)
                    print("[Info: ]x new = {}".format(get_default_x0(x_new)))
                    res_y.append(
                        pool.apply_async(self.func, args=tuple([gold_swc_tree, 
                        metric_method, get_default_x0(x_new), metric_config, optimize_config, os.path.join("tmp", "tmp_res_{}".format(j)), lock]))
                    )
                    res_x.append(get_default_x0(x_new))
                print("[Info: ]i/L = {}/{}".format(i, self.L))
                pool.close()
                pool.join()
                for it in range(len(res_x)):
                    i += 1
                    x_new_default, y_new = res_y[it].get()
                    # Metropolis
                    df = y_new - y_current
                    if df < 0 or np.exp(-df / self.T) > np.random.rand():
                        set_default_x0(x_current, x_new_default)
                        y_current = y_new
                        print("[Info: ] Jump success")
                        if y_new < self.best_y:
                            print("[Info: ] Update success")
                            self.best_x = copy.deepcopy(x_new_default)
                            self.best_y = y_new
                        break
            print("[Info: ] best x = {}".format(self.best_x))
            print("[Info: ] best y = {}".format(self.best_y))

            print("[Info: ] iter_cycle = {} T = {} stay_counter = {}".format(
                self.iter_cycle, self.T, stay_counter
            ))
            print("[Info: ]origin minimalScoreAuto = {}\n"
                  "        minimalScoreSeed = {}".format(
                self.best_x[0], self.best_x[1]
            ))
            self.iter_cycle += 1
            self.cool_down()
            self.generation_best_Y.append(self.best_y)
            self.generation_best_X.append(self.best_x)

            # if best_y stay for max_stay_counter times, stop iteration
            if self.isclose(self.best_y_history[-1], self.best_y_history[-2]):
                stay_counter += 1
            else:
                stay_counter = 0

            if self.T < self.T_min:
                stop_code = 'Cooled to final temperature'
                break
            if stay_counter > self.max_stay_counter:
                stop_code = 'Stay unchanged in the last {stay_counter} iterations'.format(stay_counter=stay_counter)
                break

        return self.best_x, self.best_y


class SAFast(SimulatedAnnealingBase):
    '''
    u ~ Uniform(0, 1, size = d)
    y = sgn(u - 0.5) * T * ((1 + 1/T)**abs(2*u - 1) - 1.0)

    xc = y * (upper - lower)
    x_new = x_old + xc

    c = n * exp(-n * quench)
    T_new = T0 * exp(-c * k**quench)
    '''
    def __init__(self, gold_swc_tree, metric_method, metric_config, optimize_config, func, **kwargs):
        # nit parent class
        super().__init__(func=func, gold_swc_tree=gold_swc_tree, metric_method=metric_method,
                         metric_config=metric_config, optimize_config=optimize_config, **kwargs)
        self.m, self.n, self.quench = kwargs.get('m', 1), kwargs.get('n', 1), kwargs.get('quench', 1)
        # upper and down are range of the parameters.
        self.lower, self.upper = kwargs.get('lower', -10), kwargs.get('upper', 10)
        self.c = self.m * np.exp(-self.n * self.quench)

    def get_new_x(self, x):
        """randomly search for a new x point"""
        x_new = copy.deepcopy(x)
        for para in x_new:
            if x_new[para]["type"] == "number":
                ran = np.random.uniform(-1, 1)
                l, r = x_new[para]['range'][0], x[para]['range'][1]
                xc = np.sign(ran) * self.T * ((1 + 1.0 / self.T) ** np.abs(ran) - 1.0)

                x_new_default = x[para]['default']/3 + xc * (r-l)/3
                x_new_default += (l+r)/3
                x_new[para]["default"] = x_new_default
            if x_new[para]["type"] == "integer":
                ran = np.random.uniform(-1, 1)
                l, r = x_new[para]['range'][0], x[para]['range'][1]
                xc = np.sign(ran) * self.T * ((1 + 1.0 / self.T) ** np.abs(ran) - 1.0)

                x_new_default = x[para]['default']/3 + xc * (r-l)/3
                x_new_default += (l+r)/3
                x_new[para]["default"] = int(np.round(x_new_default))
            if x_new[para]["type"] == "boolean":
                ran = np.random.uniform(-1, 1)
                l, r = 0, 1
                if x[para]['default'] == False:
                    x0 = 0
                else:
                    x0 = 1
                xc = np.sign(ran) * self.T * ((1 + 1.0 / self.T) ** np.abs(ran) - 1.0)

                x_new_default = x0/3 + xc * (r-l)/3
                x_new_default += (l+r)/3
                if int(np.round(x_new_default)) == 0:
                    x_new_default = False
                else:
                    x_new_default = True

                x_new[para]["default"] = x_new_default
        return x_new

    def cool_down(self):
        self.T = self.T_max * np.exp(-self.c * self.iter_cycle ** self.quench)

# SA_fast is the default
SA = SAFast



import unittest
import sys
from pyneval.metric.critical_node_metric import critical_node_metric
from pyneval.model.swc_node import SwcTree, SwcNode
from pyneval.io.read_json import read_json


class BranchDetectTest(unittest.TestCase):
    def test_branch1(self):
        sys.setrecursionlimit(1000000)
        gold_swc_tree = SwcTree()
        test_swc_tree = SwcTree()
        test_swc_tree.load("..\\..\\..\\data\\test_data\\branch_metric_data\\test\\194444.swc")
        gold_swc_tree.load("..\\..\\..\\data\\test_data\\branch_metric_data\\gold\\194444.swc")
        config = read_json("..\\..\\..\\config\\critical_node_metric.json")
        config['threshold_dis'] = 5
        config['mode'] = 1

        branch_result = critical_node_metric(test_swc_tree=gold_swc_tree, gold_swc_tree=test_swc_tree, config=config)
        print(branch_result[3])
        print(branch_result[4])
        print(branch_result[5])
        self.assertEqual(61, branch_result[3])
        self.assertEqual(73, branch_result[4])
        self.assertEqual(1.6772653444540893, branch_result[5])

    def test_branch2(self):
        sys.setrecursionlimit(1000000)
        gold_swc_tree = SwcTree()
        test_swc_tree = SwcTree()
        test_swc_tree.load("..\\..\\..\\data\\test_data\\branch_metric_data\\test\\194444.swc")
        gold_swc_tree.load("..\\..\\..\\data\\test_data\\branch_metric_data\\gold\\194444.swc")
        config = read_json("..\\..\\..\\config\\critical_node_metric.json")
        config['threshold_dis'] = 10
        branch_result = critical_node_metric(test_swc_tree=gold_swc_tree, gold_swc_tree=test_swc_tree, config=config)
        print(branch_result[3])
        print(branch_result[4])
        print(branch_result[5])
        self.assertEqual(branch_result[3], 40)
        self.assertEqual(branch_result[4], 52)
        self.assertEqual(branch_result[5], 2.335216925858577)

    def test_branch3(self):
        sys.setrecursionlimit(1000000)
        gold_swc_tree = SwcTree()
        test_swc_tree = SwcTree()
        test_swc_tree.load("..\\..\\..\\data\\test_data\\branch_metric_data\\test\\fake_data2.swc")
        gold_swc_tree.load("..\\..\\..\\data\\test_data\\branch_metric_data\\gold\\fake_data2.swc")
        config = read_json("..\\..\\..\\config\\critical_node_metric.json")
        config['threshold_dis'] = 10
        branch_result = critical_node_metric(test_swc_tree=gold_swc_tree, gold_swc_tree=test_swc_tree, config=config)

        self.assertEqual(branch_result[3], 0)
        self.assertEqual(branch_result[4], 0)
        self.assertEqual(branch_result[5], 5.031162291335031)

    def test_branch4(self):
        sys.setrecursionlimit(1000000)
        gold_swc_tree = SwcTree()
        test_swc_tree = SwcTree()
        test_swc_tree.load("..\\..\\..\\data\\test_data\\branch_metric_data\\test\\fake_data3.swc")
        gold_swc_tree.load("..\\..\\..\\data\\test_data\\branch_metric_data\\gold\\fake_data3.swc")
        config = read_json("..\\..\\..\\config\\critical_node_metric.json")
        config['threshold_dis'] = 20
        branch_result = critical_node_metric(test_swc_tree=gold_swc_tree, gold_swc_tree=test_swc_tree, config=config)
        self.assertEqual(branch_result[3], 0)
        self.assertEqual(branch_result[4], 0)
        self.assertEqual(branch_result[5], 10.44030650891055)

    def test_branch5(self):
        sys.setrecursionlimit(1000000)
        gold_swc_tree = SwcTree()
        test_swc_tree = SwcTree()
        test_swc_tree.load("..\\..\\..\\data\\test_data\\branch_metric_data\\test\\fake_data5.swc")
        gold_swc_tree.load("..\\..\\..\\data\\test_data\\branch_metric_data\\gold\\fake_data5.swc")
        config = read_json("..\\..\\..\\config\\critical_node_metric.json")
        config['threshold_dis'] = 10
        branch_result = critical_node_metric(test_swc_tree=gold_swc_tree, gold_swc_tree=test_swc_tree, config=config)

        self.assertEqual(branch_result[3], 0)
        self.assertEqual(branch_result[4], 0)
        self.assertEqual(branch_result[5], 6.073843441985439)

    def test_branch6(self):
        sys.setrecursionlimit(1000000)
        gold_swc_tree = SwcTree()
        test_swc_tree = SwcTree()
        test_swc_tree.load("..\\..\\..\\data\\test_data\\branch_metric_data\\test\\194444_isolated_node.swc")
        gold_swc_tree.load("..\\..\\..\\data\\test_data\\branch_metric_data\\gold\\194444.swc")
        config = read_json("..\\..\\..\\config\\critical_node_metric.json")
        config['threshold_dis'] = 10
        branch_result = critical_node_metric(test_swc_tree=gold_swc_tree, gold_swc_tree=test_swc_tree, config=config)
        print(branch_result[3])
        print(branch_result[4])
        print(branch_result[5])
        self.assertEqual(branch_result[3], 0)
        self.assertEqual(branch_result[4], 231)
        self.assertEqual(branch_result[5], 0.6151700217681823)

    def test_branch7(self):
        sys.setrecursionlimit(1000000)
        gold_swc_tree = SwcTree()
        test_swc_tree = SwcTree()
        test_swc_tree.load("..\\..\\..\\data\\test_data\\branch_metric_data\\test\\fake_data1.swc")
        gold_swc_tree.load("..\\..\\..\\data\\test_data\\branch_metric_data\\gold\\fake_data1.swc")
        config = read_json("..\\..\\..\\config\\critical_node_metric.json")

        branch_result = critical_node_metric(test_swc_tree=test_swc_tree,
                                             gold_swc_tree=gold_swc_tree,
                                             config=config)
        ans = [0, 1, 1, 0.0]
        for i in range(2, 6):
            print(branch_result[i])
            self.assertEqual(branch_result[i], ans[i-2])


if __name__ == '__main__':
    unittest.main()



import numpy as np
from pyneval.metric.utils import config_utils


def get_simple_lca_length(std_tree, test_gold_dict, node1, node2, switch):
    """
    get the corresponding node of node1 and node2 on std tree.
    calculate the lca distance between them
    Exception:
        Exception("[Error: ] std has not been lca initialized yet")
        std tree need to be initialized before running this function
        example:
            std_tree.get_lca_preprocess()
    """
    if std_tree.depth_array is None:
        raise Exception("[Error: ] std has not been lca initialized yet")
    std_id_node_dict = std_tree.get_id_node_dict()
    if switch:
        tmp_node1 = node1
        tmp_node2 = test_gold_dict[node2]
    else:
        tmp_node1 = test_gold_dict[node1]
        tmp_node2 = node2
    if tmp_node1 is None or tmp_node2 is None:
        raise Exception("[Error: ]gold tree and test tree are not same. ")

    lca_id = std_tree.get_lca(tmp_node1.get_id(), tmp_node2.get_id())
    if lca_id == -1:
        return config_utils.DINF
    lca_node = std_id_node_dict[lca_id]
    return tmp_node1.root_length + tmp_node2.root_length - 2 * lca_node.root_length


def get_dis_graph(gold_tree, test_tree, test_node_list, gold_node_list,
                  test_gold_dict, threshold_dis, metric_mode=1):
    """
    We use KM algorithm to get the minimum full match between gold and test branch&leaf nodes
    Since KM is used for calculating maximum match, we use the opposite value of distance
    Args:
        gold_tree(Swc_Tree):
        test_tree(Swc_Tree):
        gold_node_list(List): one list to calculate distance matrix
        test_node_list(List): the other list to calculate distance matrix
        test_gold_dict:
            corresponding relation ship between gold and test node list
        threshold_dis: if the distance of two node are larger than this threshold,
                       they are considered unlimited far
        metric_mode:
            mode = 1: distance between nodes are calculated as euclidean distance
            mode = 2: distance between nodes are calculated as distance on the gold tree
    Return:
        dis_graph(List):
            The distance matrix between nodes in gold and test node list.
            The first dimension of the graph need to be larger than the second.
        switch(bool):
            if switch is true, gold node list, test node list and corresponding length has been changed
        gold_len(integer):
        test_len(integer):
    """
    std_tree = gold_tree

    if metric_mode == 2:
        if std_tree.depth_array is None:
            std_tree.get_lca_preprocess()

    # KM works only when the length of the first dimensionality is SMALLER than the second one
    # so we need to switch gold and test when gold list is SMALLER
    switch = False
    test_len = len(test_node_list)
    gold_len = len(gold_node_list)

    if gold_len < test_len:
        switch = True
        test_len, gold_len = gold_len, test_len
        test_node_list, gold_node_list = gold_node_list, test_node_list
        gold_tree, test_tree = test_tree, gold_tree

    dis_graph = np.zeros(shape=(test_len, gold_len))

    for i in range(test_len):
        for j in range(gold_len):
            if metric_mode == 1:
                dis = test_node_list[i].distance(gold_node_list[j])
            else:
                dis = get_simple_lca_length(std_tree=std_tree,
                                            test_gold_dict=test_gold_dict,
                                            node1=test_node_list[i],
                                            node2=gold_node_list[j],
                                            switch=switch)

            if dis <= threshold_dis:
                dis_graph[i][j] = -dis
            else:
                dis_graph[i][j] = -0x3f3f3f3f/2

    dis_graph = dis_graph.tolist()
    return dis_graph, switch, test_len, gold_len


class KM:
    def __init__(self, maxn, nx, ny, G):
        self.maxn = 10000
        self.ny = ny
        self.nx = nx
        self.G = G

        self.KM_INF = 0x3f3f3f3f
        self.match = [-1]*maxn
        self.slack = [0]*maxn
        self.visx = [False]*maxn
        self.visy = [False]*maxn
        self.lx = [0]*maxn
        self.ly = [0]*maxn
        self.fa = [0]*(maxn*2)

    def find_path(self, x):
        self.visx[x] = True
        for y in range(self.ny):
            if self.visy[y]:
                continue
            tmp_delta = self.lx[x] + self.ly[y] - self.G[x][y]
            if tmp_delta < config_utils.EPS:
                self.visy[y] = True
                if self.match[y] == -1 or self.find_path(self.match[y]):
                    self.match[y] = x
                    return True
            elif self.slack[y] > tmp_delta:
                self.slack[y] = tmp_delta
        return False

    def km(self):
        for x in range(self.nx):
            for j in range(self.ny):
                self.slack[j] = self.KM_INF
            while True:
                self.visx = [False] * self.maxn
                self.visy = [False] * self.maxn
                if self.find_path(x):
                    break
                delta = self.KM_INF
                for j in range(self.ny):
                    if not self.visy[j] and delta > self.slack[j]:
                        delta = self.slack[j]
                for i in range(self.nx):
                    if self.visx[i]:
                        self.lx[i] -= delta
                for j in range(self.ny):
                    if self.visy[j]:
                        self.ly[j] += delta
                    else:
                        self.slack[j] -= delta

    def solve(self):
        self.match = [-1]*self.maxn
        self.ly = [0]*self.maxn

        for i in range(self.nx):
            self.lx[i] = -self.KM_INF
            for j in range(self.ny):
                if self.lx[i] < self.G[i][j]:
                    self.lx[i] = self.G[i][j]
        self.km()

    def get_max_dis(self):
        ans = 0.0
        for i in range(0, self.ny):
            if self.match[i] != -1 and self.G[self.match[i]][i] != -self.KM_INF/2:
                ans += self.G[self.match[i]][i]
        return ans

    def print_match(self):
        for i in range(0, self.ny):
            if self.match[i] != -1 and self.G[self.match[i]][i] != -self.KM_INF/2:
                print("gold node: {} test node {}".format(self.match[i], i))



from pyneval.io.read_json import read_json


def get_detail_type(metric_name):
    detail_type_annotation = dict()
    detail_type_annotation["ssd"] = "# 1: gold standard root\n" \
                                           "# 2: gold standard branch(degree >= 3)\n" \
                                           "# 3: gold standard continuation(degree == 2)\n" \
                                           "# 4: gold standard leaf(degree == 1)\n" \
                                           "# 5: reconstruction root\n" \
                                           "# 6: reconstruction branch(degree >= 3)\n" \
                                           "# 7: reconstruction continuation(degree == 2)\n" \
                                           "# 8: reconstruction leaf(degree == 1)\n" \
                                           "# 9: mismatched node(regardless rules above)\n\n"
    detail_type_annotation["length"] = "# 1: gold standard root\n" \
                                              "# 2: gold standard branch(degree >= 3)\n" \
                                              "# 3: gold standard continuation(degree == 2)\n" \
                                              "# 4: gold standard leaf(degree == 1)\n" \
                                              "# 5: reconstruction root\n" \
                                              "# 6: reconstruction branch(degree >= 3)\n" \
                                              "# 7: reconstruction continuation(degree == 2)\n" \
                                              "# 8: reconstruction leaf(degree == 1)\n" \
                                              "# 9: edge between this node and its parent is mismatched" \
                                              "(regardless rules above)\n\n"
    detail_type_annotation["critical_node"] = "# type of nodes in this metric detail could be change in configs\n" \
                                              "# true_positive_type: successfully reconstructed nodes, " \
                                              "exists in both GS and R\n" \
                                              "# missed: wrongly reconstructed as negative. " \
                                              "exist in GS but not in R\n" \
                                              "# excess: wrongly reconstructed as positive. " \
                                              "exist in R but not in GS.\n\n"
    detail_type_annotation["diadem"] = "# 1: gold standard root\n" \
                                              "# 2: gold standard branch(degree >= 3)\n" \
                                              "# 3: gold standard continuation(degree == 2)\n" \
                                              "# 4: gold standard leaf(degree == 1)\n" \
                                              "# 5: reconstruction root\n" \
                                              "# 6: reconstruction branch(degree >= 3)\n" \
                                              "# 7: reconstruction continuation(degree == 2)\n" \
                                              "# 8: reconstruction leaf(degree == 1)\n" \
                                              "# 9: missed: exist in GS but not in R\n" \
                                              "# 10 excess: exist in R but not in GS\n\n"
    if metric_name not in detail_type_annotation:
        return None
    return detail_type_annotation[metric_name]



import math
import os
import errno
import shutil

def get_grid_dim(x):
    """
    Transforms x into product of two integers
    :param x: int
    :return: two ints
    """
    factors = prime_powers(x)
    if len(factors) % 2 == 0:
        i = int(len(factors) / 2)
        return factors[i], factors[i - 1]

    i = len(factors) // 2
    return factors[i], factors[i]


def prime_powers(n):
    """
    Compute the factors of a positive integer
    Algorithm from https://rosettacode.org/wiki/Factors_of_an_integer#Python
    :param n: int
    :return: set
    """
    factors = set()
    for x in range(1, int(math.sqrt(n)) + 1):
        if n % x == 0:
            factors.add(int(x))
            factors.add(int(n // x))
    return sorted(factors)


def empty_dir(path):
    """
    Delete all files and folders in a directory
    :param path: string, path to directory
    :return: nothing
    """
    for the_file in os.listdir(path):
        file_path = os.path.join(path, the_file)
        try:
            if os.path.isfile(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print('Warning: {}'.format(e))


def create_dir(path):
    """
    Creates a directory
    :param path: string
    :return: nothing
    """
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise


def prepare_dir(path, empty=False):
    """
    Creates a directory if it soes not exist
    :param path: string, path to desired directory
    :param empty: boolean, delete all directory content if it exists
    :return: nothing
    """
    if not os.path.exists(path):
        create_dir(path)

    if empty:
        empty_dir(path)



import tflib as lib

import numpy as np
import tensorflow as tf

def Batchnorm(name, axes, inputs, is_training=None, stats_iter=None, update_moving_stats=True, fused=True, labels=None, n_labels=None):
    """conditional batchnorm (dumoulin et al 2016) for BCHW conv filtermaps"""
    if axes != [0,2,3]:
        raise Exception('unsupported')
    mean, var = tf.nn.moments(inputs, axes, keep_dims=True)
    shape = mean.get_shape().as_list() # shape is [1,n,1,1]
    offset_m = lib.param(name+'.offset', np.zeros([n_labels,shape[1]], dtype='float32'))
    scale_m = lib.param(name+'.scale', np.ones([n_labels,shape[1]], dtype='float32'))
    offset = tf.nn.embedding_lookup(offset_m, labels)
    scale = tf.nn.embedding_lookup(scale_m, labels)
    result = tf.nn.batch_normalization(inputs, mean, var, offset[:,:,None,None], scale[:,:,None,None], 1e-5)
    return result


import unittest
from pyneval.metric.utils.edge_match_utils import get_lca_length, get_edge_rtree, get_idedge_dict
from pyneval.model.swc_node import SwcTree, SwcNode
from pyneval.model.euclidean_point import EuclideanPoint, Line


class TestGetLcaLength(unittest.TestCase):
    test_swc_tree = SwcTree()
    test_rtree = None
    id_edge_dict = None

    def init(self):
        self.test_swc_tree.load("../../data_example/unit_test/get_nearby_edges_tree1.swc")
        self.test_rtree = get_edge_rtree(self.test_swc_tree)
        self.id_edge_dict = get_idedge_dict(self.test_swc_tree)

        self.test_swc_tree.get_lca_preprocess()

    def test_1(self):
        self.init()
        line_tuple_a = tuple([self.test_swc_tree.node_from_id(8),
                              self.test_swc_tree.node_from_id(7)])
        line_tuple_b = tuple([self.test_swc_tree.node_from_id(7),
                              self.test_swc_tree.node_from_id(6)])
        e_node1 = EuclideanPoint(center=[1.03, -2.1, 6.31])
        e_node2 = EuclideanPoint(center=[3.87, 0.98, 1.17])

        test_length = get_lca_length(self.test_swc_tree, \
                                     line_tuple_a, \
                                     line_tuple_b, \
                                     Line(e_node_1=e_node1,
                                          e_node_2=e_node2))
        self.assertEqual(test_length, 9.589854552695694)

    def test_2(self):
        self.init()
        line_tuple_a = tuple([self.test_swc_tree.node_from_id(4),
                              self.test_swc_tree.node_from_id(3)])
        line_tuple_b = tuple([self.test_swc_tree.node_from_id(15),
                              self.test_swc_tree.node_from_id(9)])
        e_node1 = EuclideanPoint(center=[-1.89657, 6.51822, -1.40403])
        e_node2 = EuclideanPoint(center=[-2.02446, 0.54277, 7.48183])

        test_length = get_lca_length(self.test_swc_tree, \
                                     line_tuple_a, \
                                     line_tuple_b, \
                                     Line(e_node_1=e_node1,
                                          e_node_2=e_node2))
        self.assertEqual(test_length, 34.372721303735716)


if __name__ == '__main__':
    unittest.main()



import cProfile
import pstats
import os


# 性能分析装饰器定义
def do_cprofile(filename):
    """
    Decorator for function profiling.
    """
    def wrapper(func):
        def profiled_func(*args, **kwargs):
            # Flag for do profiling or not.
            DO_PROF = False
            if DO_PROF:
                profile = cProfile.Profile()
                profile.enable()
                result = func(*args, **kwargs)
                profile.disable()
                # Sort stat by internal time.
                sortby = "tottime"
                ps = pstats.Stats(profile).sort_stats(sortby)
                ps.dump_stats(filename)
            else:
                result = func(*args, **kwargs)
            return result
        return profiled_func
    return wrapper


# USAGE
# python sliding_window.py --image images/adrian_florida.jpg 

# import the necessary packages
from pyimagesearch.helpers import pyramid
from pyimagesearch.helpers import sliding_window
import argparse
import time
import cv2

# construct the argument parser and parse the arguments
ap = argparse.ArgumentParser()
ap.add_argument("-i", "--image", required=True, help="Path to the image")
args = vars(ap.parse_args())

# load the image and define the window width and height
image = cv2.imread(args["image"])
(winW, winH) = (128, 128)

# loop over the image pyramid
for resized in pyramid(image, scale=1.5):
	# loop over the sliding window for each layer of the pyramid
	for (x, y, window) in sliding_window(resized, stepSize=32, windowSize=(winW, winH)):
		# if the window does not meet our desired window size, ignore it
		if window.shape[0] != winH or window.shape[1] != winW:
			continue

		# THIS IS WHERE YOU WOULD PROCESS YOUR WINDOW, SUCH AS APPLYING A
		# MACHINE LEARNING CLASSIFIER TO CLASSIFY THE CONTENTS OF THE
		# WINDOW

		# since we do not have a classifier, we'll just draw the window
		clone = resized.copy()
		cv2.rectangle(clone, (x, y), (x + winW, y + winH), (0, 255, 0), 2)
		cv2.imshow("Window", clone)
		cv2.waitKey(1)
		time.sleep(0.025)


"""Example Google style docstrings.

This module define the labels for each class images.

"""

label_d_class = {
    0:'branch',
    1:'nonbranch'
}

class_d_label = {
    'branch':0,
    'nonbranch':1
}

class_d_name = {
    'branch':2,
    'nonbranch':1
}

class_names = list(class_d_label.keys())

label_class_names = class_names.copy()
label_class_names.append('none')
LabelPatchSizeRestrict = (32,32)
PlotLabelDir = '/home/guanwanxian/LearnPython/NeuTube/TrainModel/Extra/PlotLabel32by32'
MouseSampleDir = '/home/guanwanxian/LearnPython/NeuTube/TrainModel/Extra/MouseSample32by32'
MouseLabelDir = '/home/guanwanxian/LearnPython/NeuTube/TrainModel/Extra/MouseLabel32by32'



# @Author: guanwanxian
# @Date:   1970-01-01T08:00:00+08:00
# @Email:  guanwanxian@zju.edu.cn
# @Last modified by:   guanwanxian
# @Last modified time: 2017-02-25T21:24:37+08:00

import numpy as np

def rotMatrix3d(theta, axis):
    '''rotMatrix3d: Given axis and theta, calculate rotation matrix in 3d axes.
    Args:
        theta (int): The negative angle along  right-hand screw rule.
        axis (string): The axis which rotate along with.
            'x','y' or 'z'
    Returns:
        M (numpy array): The rotation matrix.
    '''
    assert axis in ['x','y','z']
    c, s = np.cos(theta/180*np.pi), np.sin(theta/180*np.pi)
    if axis=='x':
        M = np.matrix([[1, 0, 0, 0],
                       [0, c, -s,0],
                       [0, s, c, 0],
                       [0, 0, 0, 1]])
    elif axis=='y':
        M = np.matrix([[c, 0, s, 0],
                       [0, 1, 0,  0],
                       [-s,0, c,  0],
                       [0, 0, 0, 1]])
    else: # 'z'
        M = np.matrix([[c, -s, 0, 0],
                       [s, c,  0,0],
                       [0, 0, 1, 0],
                       [0, 0, 0, 1]])
    return M

def transMatrix3d(tx, ty, tz):
    M = np.matrix([[1,0,0,tx],
                   [0,1,0,ty],
                   [0,0,1,tz],
                   [0,0,0,1]])
    return M

def transMatrix3dCompose(theta_z, theta_y, tx, ty, tz):
    m_r_z = rotMatrix3d(theta_z, 'z')
    m_r_y = rotMatrix3d(theta_y, 'y')
    m_t = transMatrix3d(tx, ty, tz)
    return reduce(np.matmul, [m_t, m_r_y, m_r_z])

def appendMatrix3d(m, m_add):
    return np.matmul(m, m_add)

def cosAngle(x):
    return np.cos(x/180*np.pi)

def sinAngle(x):
    return np.sin(x/180*np.pi)



import math
import glob
import numpy as np
from . import TiffFile

def open_swc(file_name):
	return SWC(file_name)
	
def open_tif(file_name):
	'''
		return numpy array
		z,y,x
	'''
	return TiffFile.imread(file_name)
	
def save_tif(np_array_like,file_name):
	'''
		save numpy array like to tiff 
		z,y,x
	'''
	TiffFile.imsave(file_name,np_array_like.astype(np.uint8))
	
def save_swc2tif(swc_file_name,tif_file_name,shape=None):
	swc=open_swc(swc_file_name)
	bound_box=swc.get_bound_box()
	width=math.ceil(bound_box[3]-bound_box[0])
	height=math.ceil(bound_box[4]-bound_box[1])
	depth=math.ceil(bound_box[5]-bound_box[2])
	if shape:
	    depth,height,width=shape
	tif=np.zeros(shape=(depth,height,width))
	for node in swc.nodes:
		_sphere(tif,node.x,node.y,node.z,node.r,255)
	for edge in swc.edges:
		if edge[0]!=-1 and edge[1]!=-1:
			a,b=swc.get_node(edge[0]),swc.get_node(edge[1])
			_cone(tif,a.x,a.y,a.z,a.r,b.x,b.y,b.z,b.r,255)
	save_tif(tif,tif_file_name)
	
class SWCNode(object):
	
	def __init__(self,id,type,x,y,z,r,pid):
		self.id=id
		self.type=type
		self.x=x
		self.y=y
		self.z=z
		self.r=r
		self.pid=pid
		
class SWC(object):
	
	def __init__(self,file_name=None):
		self.__nodes={}
		self.__edges=[]
		self.bound_box=[0,0,0,0,0,0]# x0,y0,z0,x1,y1,z1
		if file_name:
			self.open(file_name)
			
	def open(self,file_name):
		with open(file_name) as f:
			for line in f.readlines():
				if line.startswith('#'):
					continue
				id,type,x,y,z,r,pid=map(float,line.split())
				self.__nodes[id]=SWCNode(id,type,x,y,z,r,pid)
				self.__edges.append((id,pid))
				if x<self.bound_box[0]:
					self.bound_box[0]=x
				if x>self.bound_box[3]:
					self.bound_box[3]=x
				if y<self.bound_box[1]:
					self.bound_box[1]=y
				if y>self.bound_box[4]:
					self.bound_box[4]=y
				if z<self.bound_box[2]:
					self.bound_box[2]=z
				if z>self.bound_box[5]:
					self.bound_box[5]=z
					
	def get_node(self,id):
		return self.__nodes[id]
		
	def get_bound_box(self):
		return self.bound_box
		
	@property
	def nodes(self):
		return self.__nodes.values()
	
	@property
	def edges(self):
		return self.__edges
	
def _sphere(stack,x,y,z,r,v=1):
	d,h,w=stack.shape
	for k in range(int(max(0,z-r)),int(min(d,z+r))):
		for j in range(int(max(0,y-r)),int(min(h,y+r))):
			for i in range(int(max(0,x-r)),int(min(w,x+r))):
				if (k-z)**2+(j-y)**2+(i-x)**2<=r**2:
					stack[k,j,i]=v
	
def _cone(stack,x0,y0,z0,r0,x1,y1,z1,r1,v=1):
	d,h,w=stack.shape
	'''if r0<r1:
		x0,x1=x1,x0
		y0,y1=y1,y0
		z0,z1=z1,z0
		r0,r1=r1,r0'''
	a,b,c=x1-x0,y1-y0,z1-z0
	r=math.sqrt(a**2+b**2+c**2)+1e-6
	tan_theta=(r0-r1)/r
	points=[]
	for k in range(math.ceil(r)):
		rk=tan_theta*(r-k)+r1
		for j in range(-int(rk)-1,int(rk)+1):
			for i in range(-int(rk)-1,int(rk)+1):
				if j**2+i**2<rk**2:
					points.append((i,j,k))
	n1=math.sqrt(b**2+c**2)
	if n1==0:
		for pt in points:
			x=int(pt[2]+x0)
			y=int(pt[1]+y0)
			z=int(pt[0]+z0)
			if x>=0 and x<w and y>=0 and y<h and z>=0 and z<d:
				stack[z,y,x]=v
	else:	
		n2=math.sqrt((b**2+c**2)**2+(a*b)**2+(a*c)**2)
		Tr=np.array(((0,c/n1,-b/n1),(-(b**2+c**2)/n2,a*b/n2,a*c/n2),(a/r,b/r,c/r)))
		Tr=Tr.T
		for pt in points:
			coord=np.matmul(Tr,pt)
			coord=coord+(x0,y0,z0)
			for x in range(math.floor(coord[0]),math.ceil(coord[0])+1):
				for y in range(math.floor(coord[1]),math.ceil(coord[1])+1):
					for z in range(math.floor(coord[2]),math.ceil(coord[2])+1):
						if x>=0 and x<w and y>=0 and y<h and z>=0 and z<d:
							stack[z,y,x]=v
	
if __name__=='__main__':
    for swc_file in glob.glob('./*.swc'):
	    save_swc2tif(swc_file,swc_file+'.tif',(85,2048,2048))



from PIL import Image
from PIL import ImageDraw
import numpy as np
import matplotlib.pyplot as plt
from Utils import joinPath, saveData, setDir
from PlotTools import plotMyFigureOnAxes

def extractSelectedRegion(selected_data):
    img_array = selected_data['image_array']
    W_R = int(selected_data['W_R'])
    H_R = int(selected_data['H_R'])
    # import pdb
    # pdb.set_trace()
    return np.array([img_array[x[1]-H_R:x[1]+H_R, x[0]-W_R:x[0]+W_R] for x in selected_data['selection']['regions']])

def redrawFromSelectedData(ax, selected_data, circle_R=3):
    redraw(ax, Image.fromarray(selected_data['image_array']),
                               selected_data['selection'],
                               selected_data['W_R'],
                               selected_data['H_R'],
                               circle_R)


def redraw(ax, img, selection, W_R, H_R, circle_R):
    """Redraw image presented in numpy format(img) on Matplotlib Axes(ax)
       :ax : Matplotlib axes, where to draw image
       :img : Pillow Image Object
       :selection: {'regions':[], 'count':int}, regions contain the regions to draw
       :W_R : int, half of window width
       :H_R : int, half of window height
    """
    ax.cla()
    img_rgb = img.convert(mode='RGB')
    draw = ImageDraw.Draw(img_rgb)

    for idx, lb in enumerate(selection['regions']):
        if lb[2]==0: # branch
            # The box is a 4-tuple defining the left, upper, right, and lower pixel coordinate
            draw.rectangle([lb[0]-W_R, lb[1]-H_R, lb[0]+W_R, lb[1]+H_R]
                            ,outline="red")

            draw.ellipse([lb[0]-circle_R, lb[1]-circle_R, lb[0]+circle_R, lb[1]+circle_R]
                         , fill="red", outline=None) # "red"
            draw.text((lb[0]-W_R,lb[1]-H_R), '{0}'.format(idx), fill='yellow')
        else: # nonbranch
            draw.rectangle([lb[0]-W_R, lb[1]-H_R, lb[0]+W_R, lb[1]+H_R]
                                ,outline="green")
            draw.ellipse([lb[0]-circle_R, lb[1]-circle_R, lb[0]+circle_R, lb[1]+circle_R]
                         , fill="green", outline=None) # "green"
            draw.text((lb[0]-W_R,lb[1]-H_R), '{0}'.format(idx), fill='yellow')

    del draw

    plotMyFigureOnAxes(ax, np.asarray(img_rgb))
    # im = ax.imshow()
    # fig = ax.figure
    # fig.subplots_adjust(right=0.8)
    # cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
    # fig.colorbar(im, cax=cbar_ax)

def checkBoxInside(img_size, x, y, box_w, box_h):
    '''checkBoxInside: Check whether mouse selected region inside
    Args:
        img_size (tuple(int,int)): image size (width, height).
        x (int): x position of box.
        y (int): y position of box.
        box_w (int): width of box.
        box_h (int): height of box.

    Returns:
        bool: Whether box is inside this image.
    '''
    left = x-box_w
    right = x+box_w
    up = y-box_h
    down = y+box_h
    return left>=0 and right<=img_size[0] and up>=0 and down<=img_size[1]


def mouseSample(dst_pckl_path, src_dir, img_name, W_R, H_R, circle_R):
    """mouseSample, use mouse to select [2*W_R, 2*H_R] windows in images
       :dst_pckl_path : string, destination path to store extracted variables
       :src_dir : string, source dir of image
       :img_name : string, image name
       :W_R : int, half window width
       :H_R : int , half window height
       :circle_R : int, radiu of circle drawn on image
    """
    setDir(dst_pckl_path)
    img_path = joinPath(src_dir, img_name)

    img = Image.open(img_path)
    print('Image size: {}'.format(img.size))
    img_size = img.size
    img_array = np.asarray(img)

    selection = {'regions':[], 'count':0}

    # fig_a and fig_b are created differently for each different image
    # and it's okay they are load simultaneously
    fig_a = plt.figure()
    fig_a.clf()
    fig_a_ax = fig_a.add_subplot(111)
    plotMyFigureOnAxes(fig_a_ax, img_array)
    # fig_a_ax.imshow(img_array, cmap='Greys_r')

    fig_b = plt.figure()
    fig_b.clf()
    fig_b_ax = fig_b.add_subplot(111)

    cids = []

    def onclick(event, selection, W_R, H_R, circle_R):
        center_x = int(event.xdata)
        center_y = int(event.ydata)
        fig_b_ax.cla()
        if checkBoxInside(img_size, center_x, center_y, W_R, H_R):
            window = img_array[center_y-H_R:center_y+H_R, center_x-W_R:center_x+W_R]
            window_img = Image.fromarray(window)
            window_arr = np.asarray(window_img)
            plotMyFigureOnAxes(fig_b_ax, window_arr)
            # fig_b_ax.imshow(window_arr,cmap='Greys_r')

            fig_b_ax.text(0,0.4,'img{0}, ({1},{2})'.format(selection['count'], center_x, center_y), color='r')
            selection['count'] += 1
            selection['regions'].append((center_x, center_y, 0))

            redraw(fig_a_ax, img, selection, W_R, H_R, circle_R)
        else:
            fig_b_ax.text(0,0.4,'The box is outside of image', color='r')

    def on_key(event, ax, cids):
        if event.key=='ctrl+alt+n':
            for cid in cids:
                ax.figure.canvas.mpl_disconnect(cid)

            # Save data into a pickl file
            selected_data = {}
            selected_data['selection'] = selection
            selected_data['image_array'] = img_array
            selected_data['img_name'] = img_name
            selected_data['W_R'] = W_R
            selected_data['H_R'] = H_R

            save_pickl_file_name = '_'.join(img_name.split('.')[:-1]).replace('/', '_')
            save_pickl_file_name += '.pckl'

            print('saved pickl file name: {}'.format(save_pickl_file_name))
            saveData([selected_data], joinPath(dst_pckl_path, save_pickl_file_name))

            # Notify
            fig_a_ax.text(0,0.4,'Finishe mouse selection', color='r')

    # Register event handler
    cids.append(fig_a_ax.figure.canvas.mpl_connect('button_press_event', lambda event: onclick(event, selection, W_R, H_R, circle_R)))
    cids.append(fig_a_ax.figure.canvas.mpl_connect('key_press_event', lambda event: on_key(event, fig_a_ax, cids)))



"""Converts images to TFRecords file format with Example protos."""

import os
import tensorflow as tf

def _int64_feature(value):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))


def _bytes_feature(value):
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def convertImagesToTFRecordsFile(images, labels, shapes, directorys, name):
    num_examples = labels.shape[0]
    print('labels shape is: ', labels.shape[0])
    if num_examples != images.shape[0] != labels.shape[0]:
        raise ValueError("Images size %d does not match label size %d." %
                                         (images.shape[0], num_examples))

    filename = os.path.join(directorys, name + '.tfrecords')
    print('Writing', filename)
    writer = tf.python_io.TFRecordWriter(filename)
    for index in range(num_examples):
        image_raw = images[index].tostring()
        image_shape = shapes[index]
        example = tf.train.Example(features=tf.train.Features(feature={
                'height': _int64_feature(int(image_shape[0])),
                'width': _int64_feature(int(image_shape[1])),
                'depth': _int64_feature(int(image_shape[2])),
                'label': _int64_feature(int(labels[index])),
                'image_raw': _bytes_feature(image_raw)}))
        writer.write(example.SerializeToString())

def convertImagesToTFRecordsFile3D(images, labels, shapes, directorys, name):
    num_examples = labels.shape[0]
    print('labels shape is: ', labels.shape[0])
    if num_examples != images.shape[0] != labels.shape[0]:
        raise ValueError("Images size %d does not match label size %d." %
                                         (images.shape[0], num_examples))

    filename = os.path.join(directorys, name + '.tfrecords')
    print('Writing', filename)
    writer = tf.python_io.TFRecordWriter(filename)
    for index in range(num_examples):
        image_raw = images[index].tostring()
        image_shape = shapes[index]
        example = tf.train.Example(features=tf.train.Features(feature={
                'depth': _int64_feature(int(image_shape[0])),
                'width': _int64_feature(int(image_shape[1])),
                'height': _int64_feature(int(image_shape[2])),
                'channel': _int64_feature(int(image_shape[3])),
                'label': _int64_feature(int(labels[index])),
                'image_raw': _bytes_feature(image_raw)}))
        writer.write(example.SerializeToString())

if __name__ == '__main__':
    tf.app.run()



import tflib as lib

import numpy as np
import tensorflow as tf

_default_weightnorm = False
def enable_default_weightnorm():
    global _default_weightnorm
    _default_weightnorm = True

def disable_default_weightnorm():
    global _default_weightnorm
    _default_weightnorm = False

_weights_stdev = None
def set_weights_stdev(weights_stdev):
    global _weights_stdev
    _weights_stdev = weights_stdev

def unset_weights_stdev():
    global _weights_stdev
    _weights_stdev = None

def Linear(
        name,
        input_dim,
        output_dim,
        inputs,
        biases=True,
        initialization=None,
        weightnorm=None,
        gain=1.
        ):
    """
    initialization: None, `lecun`, 'glorot', `he`, 'glorot_he', `orthogonal`, `("uniform", range)`
    """
    with tf.name_scope(name) as scope:

        def uniform(stdev, size):
            if _weights_stdev is not None:
                stdev = _weights_stdev
            return np.random.uniform(
                low=-stdev * np.sqrt(3),
                high=stdev * np.sqrt(3),
                size=size
            ).astype('float32')

        if initialization == 'lecun':# and input_dim != output_dim):
            # disabling orth. init for now because it's too slow
            weight_values = uniform(
                np.sqrt(1./input_dim),
                (input_dim, output_dim)
            )

        elif initialization == 'glorot' or (initialization == None):

            weight_values = uniform(
                np.sqrt(2./(input_dim+output_dim)),
                (input_dim, output_dim)
            )

        elif initialization == 'he':

            weight_values = uniform(
                np.sqrt(2./input_dim),
                (input_dim, output_dim)
            )

        elif initialization == 'glorot_he':

            weight_values = uniform(
                np.sqrt(4./(input_dim+output_dim)),
                (input_dim, output_dim)
            )

        elif initialization == 'orthogonal' or \
            (initialization == None and input_dim == output_dim):

            # From lasagne
            def sample(shape):
                if len(shape) < 2:
                    raise RuntimeError("Only shapes of length 2 or more are "
                                       "supported.")
                flat_shape = (shape[0], np.prod(shape[1:]))
                 # TODO: why normal and not uniform?
                a = np.random.normal(0.0, 1.0, flat_shape)
                u, _, v = np.linalg.svd(a, full_matrices=False)
                # pick the one with the correct shape
                q = u if u.shape == flat_shape else v
                q = q.reshape(shape)
                return q.astype('float32')
            weight_values = sample((input_dim, output_dim))

        elif initialization[0] == 'uniform':

            weight_values = np.random.uniform(
                low=-initialization[1],
                high=initialization[1],
                size=(input_dim, output_dim)
            ).astype('float32')

        else:

            raise Exception('Invalid initialization!')

        weight_values *= gain

        weight = lib.param(
            name + '.W',
            weight_values
        )

        if weightnorm==None:
            weightnorm = _default_weightnorm
        if weightnorm:
            norm_values = np.sqrt(np.sum(np.square(weight_values), axis=0))
            # norm_values = np.linalg.norm(weight_values, axis=0)

            target_norms = lib.param(
                name + '.g',
                norm_values
            )

            with tf.name_scope('weightnorm') as scope:
                norms = tf.sqrt(tf.reduce_sum(tf.square(weight), reduction_indices=[0]))
                weight = weight * (target_norms / norms)

        # if 'Discriminator' in name:
        #     print "WARNING weight constraint on {}".format(name)
        #     weight = tf.nn.softsign(10.*weight)*.1

        if inputs.get_shape().ndims == 2:
            result = tf.matmul(inputs, weight)
        else:
            reshaped_inputs = tf.reshape(inputs, [-1, input_dim])
            result = tf.matmul(reshaped_inputs, weight)
            result = tf.reshape(result, tf.pack(tf.unpack(tf.shape(inputs))[:-1] + [output_dim]))

        if biases:
            result = tf.nn.bias_add(
                result,
                lib.param(
                    name + '.b',
                    np.zeros((output_dim,), dtype='float32')
                )
            )

        return result



import tflib as lib

import numpy as np
import tensorflow as tf

_default_weightnorm = False
def enable_default_weightnorm():
    global _default_weightnorm
    _default_weightnorm = True

_weights_stdev = None
def set_weights_stdev(weights_stdev):
    global _weights_stdev
    _weights_stdev = weights_stdev

def unset_weights_stdev():
    global _weights_stdev
    _weights_stdev = None

def Deconv3D(
    name,
    input_dim,
    output_dim,
    filter_size,
    inputs,
    he_init=True,
    weightnorm=None,
    biases=True,
    gain=1.,
    mask_type=None,
    ):
    """
    inputs: tensor of shape (batch size, depth, height, width, input_dim)
    returns: tensor of shape (batch size, depth, height, width, output_dim)
    """
    with tf.name_scope(name) as scope:

        if mask_type != None:
            raise Exception('Unsupported configuration')

        def uniform(stdev, size):
            return np.random.uniform(
                low=-stdev * np.sqrt(3),
                high=stdev * np.sqrt(3),
                size=size
            ).astype('float32')

        stride = 1
        fan_in = input_dim * filter_size**2 / (stride**2)
        fan_out = output_dim * filter_size**2

        if he_init:
            filters_stdev = np.sqrt(4./(fan_in+fan_out))
        else: # Normalized init (Glorot & Bengio)
            filters_stdev = np.sqrt(2./(fan_in+fan_out))


        if _weights_stdev is not None:
            filter_values = uniform(
                _weights_stdev,
                (filter_size, filter_size, filter_size, output_dim, input_dim)
            )
        else:
            filter_values = uniform(
                filters_stdev,
                (filter_size, filter_size, filter_size, output_dim, input_dim)
            )

        filter_values *= gain

        filters = lib.param(
            name+'.Filters',
            filter_values
        )

        if weightnorm==None:
            weightnorm = _default_weightnorm
        if weightnorm:
            norm_values = np.sqrt(np.sum(np.square(filter_values), axis=(0,1,2,4)))
            target_norms = lib.param(
                name + '.g',
                norm_values
            )
            with tf.name_scope('weightnorm') as scope:
                norms = tf.sqrt(tf.reduce_sum(tf.square(filters), reduction_indices=[0,1,2,4]))
                filters = filters * tf.expand_dims(target_norms / norms, 1)


        inputs = tf.transpose(inputs, [0,2,3,4,1], name='NCDHW_to_NDHWC')

        input_shape = tf.shape(inputs)
        try: # tf pre-1.0 (top) vs 1.0 (bottom)
            output_shape = tf.pack([input_shape[0], input_shape[1], input_shape[2], input_shape[3], output_dim])
        except Exception as e:
            output_shape = tf.stack([input_shape[0], input_shape[1], input_shape[2], input_shape[3], output_dim])

        result = tf.nn.conv3d_transpose(
            value=inputs,
            filter=filters,
            output_shape=output_shape,
            strides=[1, 1, 1, 1, 1],
            padding='SAME'
        )

        if biases:
            _biases = lib.param(
                name+'.Biases',
                np.zeros(output_dim, dtype='float32')
            )
            result = tf.nn.bias_add(result, _biases)

        result = tf.transpose(result, [0,4,1,2,3], name='NDHWC_to_NCDHW')


        return result



import numpy as np

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

import collections
import time
import pickle

_since_beginning = collections.defaultdict(lambda: {})
_since_last_flush = collections.defaultdict(lambda: {})

_iter = [0]

def setIter(val):
	_iter[0] = val

def tick():
	_iter[0] += 1

def plot(name, value):
	_since_last_flush[name][_iter[0]] = value

def flush():
	prints = []

	for name, vals in _since_last_flush.items():
		prints.append("{}\t{}".format(name, np.mean(list(vals.values()))))
		_since_beginning[name].update(vals)

		x_vals = np.sort(list(_since_beginning[name].keys()))
		y_vals = [_since_beginning[name][x] for x in x_vals]

		plt.clf()
		plt.plot(x_vals, y_vals)
		plt.xlabel('iteration')
		plt.ylabel(name)
		plt.savefig(name.replace(' ', '_')+'.jpg')

	print("iter {}\t{}".format(_iter[0], "\t".join(prints)))
	_since_last_flush.clear()

	with open('log.pkl', 'wb') as f:
		pickle.dump(dict(_since_beginning), f, pickle.HIGHEST_PROTOCOL)



import unittest
from pyneval.metric.utils.edge_match_utils import get_nearby_edges, get_edge_rtree, get_idedge_dict
from pyneval.model.swc_node import SwcTree, SwcNode
from pyneval.model.euclidean_point import EuclideanPoint


class TestGetNearbyEdges(unittest.TestCase):
    test_swc_tree = SwcTree()
    test_rtree = None
    id_edge_dict = None

    def init(self):
        self.test_swc_tree.load("../../data_example/unit_test/get_nearby_edges_tree1.swc")
        self.test_rtree = get_edge_rtree(self.test_swc_tree)
        self.id_edge_dict = get_idedge_dict(self.test_swc_tree)

    def test_1(self):
        line = [[7, 6], [8, 7]]
        dis = [0.009317785645644346, 0.07576191838507501]

        self.init()
        e_point = EuclideanPoint(center=[5, -2.7, 4.8])
        point = SwcNode(center=e_point)
        threshold = 0.1

        res = get_nearby_edges(self.test_rtree, point, self.id_edge_dict, threshold, False, DEBUG=False)
        self.assertEqual(len(res), 2)

        for i in range(len(res)):
            element = res[i]
            self.assertEqual(element[0][0].get_id(), line[i][0])
            self.assertEqual(element[0][1].get_id(), line[i][1])
            self.assertEqual(element[1], dis[i])

    def test_2(self):
        line = [[7, 6], [8, 7]]
        dis = [0.009317785645644346, 0.07576191838507501]

        self.init()
        e_point = EuclideanPoint(center=[5.02439, -2.75148, 4.84998])
        point = SwcNode(center=e_point)
        threshold = 0.1

        res = get_nearby_edges(self.test_rtree, point, self.id_edge_dict, threshold, True, DEBUG=False)
        print(res.__str__())
        self.assertEqual(len(res), 0)


if __name__ == "__main__":
    unittest.main()





import os
from functools import reduce
from collections import  deque
import numpy as np
import scipy as sp
from numpy import linalg as LA
from scipy.spatial import distance_matrix
from .Transformations import rotation_matrix, superimposition_matrix
from .SWCExtractor import Vertex
from .Obj3D import Point3D, Sphere, Cone, calculateBound, calScaleRatio
from .Utils import  Timer
from . import Draw3DTools
from .Draw3DTools import randIntList
from . import ImageUtils

def getRandChildNumber():
    ''' Random generate children number of a tree node
        Input:
            None
        Output:
            (int) : Children number
    '''
    return np.random.choice([1,2,3,4], p=[0.5, 0.35, 0.1, 0.05])



def getChildRadius(base_x, depth, max_depth):
    if depth==0: # root
        return np.random.choice([base_x*3,base_x*5,base_x*10], p=[0.25,0.5,0.25])
    else:
        return np.random.choice([base_x*1,base_x*2,base_x*3,base_x*4,base_x*5], p=[0.2, 0.2, 0.2, 0.2, 0.2])



def getChildLength(base_length, depth, max_depth):
    ''' 子节点距离父节点的长度
    '''
    return base_length + (max_depth-depth) + np.random.randint(0,base_length//2)



def setMarkWithSphere(mark, sphere, mark_shape, lower, upper, use_bbox=False):
    bbox = list(sphere.calBBox()) # xmin,ymin,zmin,xmax,ymax,zmax
    for i in range(3):
        j = i+3
        if (bbox[i]<0):
            bbox[i] = 0
        if (bbox[j]>mark_shape[i]):
            bbox[j] = mark_shape[i]
    (xmin,ymin,zmin,xmax,ymax,zmax) = tuple(bbox)
    (x_idxs,y_idxs,z_idxs)=np.where(mark[xmin:xmax,ymin:ymax,zmin:zmax]==0)
    # points=img_idxs[:3, xmin+x_idxs, ymin+y_idxs, zmin+z_idxs] # 3*M
    # points=points.T # M*3
    if not use_bbox:
        xs = np.asarray(xmin+x_idxs).reshape((len(x_idxs),1))
        ys = np.asarray(ymin+y_idxs).reshape((len(y_idxs),1))
        zs = np.asarray(zmin+z_idxs).reshape((len(z_idxs),1))
        points=np.hstack((xs,ys,zs))

        sphere_c_mat = np.array([sphere.center_point.toList()]) # 1*3
        # 计算所有点到所有球心的距离
        dis_mat = distance_matrix(points,sphere_c_mat) # M*1

        # 判断距离是否小于半径
        res_idxs = np.where(dis_mat<=sphere.radius)[0]
        # value_list = randIntList(lower,upper,len(res_idxs))
        for pos in res_idxs:
            value = np.random.randint(lower, upper);
            mark[xmin+x_idxs[pos], ymin+y_idxs[pos], zmin+z_idxs[pos]] = value
        # mark[xmin+x_idxs[res_idxs], ymin+y_idxs[res_idxs], zmin+z_idxs[res_idxs]] = 255
    else:
        # value_list = randIntList(lower,upper,len(res_idxs))
        for (px,py,pz) in zip(x_idxs,y_idxs,z_idxs):
            value = np.random.randint(lower, upper);
            mark[xmin+x_idxs[px], ymin+y_idxs[py], zmin+z_idxs[pz]] = value
        # mark[xmin+x_idxs, ymin+y_idxs, zmin+z_idxs] = 255



def setMarkWithCone(mark, cone, mark_shape, lower, upper, use_bbox=False):
    bbox = list(cone.calBBox()) # xmin,ymin,zmin,xmax,ymax,zmax
    for i in range(3):
        j = i+3
        if (bbox[i]<0):
            bbox[i] = 0
        if (bbox[j]>mark_shape[i]):
            bbox[j] = mark_shape[i]
    (xmin,ymin,zmin,xmax,ymax,zmax) = tuple(bbox)

    (x_idxs,y_idxs,z_idxs)=np.where(mark[xmin:xmax,ymin:ymax,zmin:zmax]==0)
    if not use_bbox:
        xs = np.asarray(xmin+x_idxs).reshape((len(x_idxs),1))
        ys = np.asarray(ymin+y_idxs).reshape((len(y_idxs),1))
        zs = np.asarray(zmin+z_idxs).reshape((len(z_idxs),1))
        ns = np.ones((len(z_idxs),1))
        points=np.hstack((xs,ys,zs,ns))

        # 每个圆锥的还原矩阵
        r_min=cone.up_radius
        r_max=cone.bottom_radius
        height=cone.height
        cone_revert_mat = cone.revertMat().T # 4*4

        # 每个椎体还原后坐标
        revert_coor_mat = np.matmul(points, cone_revert_mat) # M*4
        revert_radius_list = LA.norm(revert_coor_mat[:,:2], axis=1) # M

        # Local Indexs
        M = points.shape[0]
        l_idx = np.arange(M) # M (1-dim)
        l_mark = np.ones((M,), dtype=bool)

        # 过滤高度在外部的点
        res_idxs = np.logical_or(revert_coor_mat[l_idx[l_mark],2]<0, revert_coor_mat[l_idx[l_mark],2]>height)
        l_mark[l_idx[l_mark][res_idxs]]=False

        # 过滤半径在外部的点
        res_idxs = revert_radius_list[l_idx[l_mark]]>r_max
        l_mark[l_idx[l_mark][res_idxs]]=False

        # 过滤半径在内部的点
        res_idxs = revert_radius_list[l_idx[l_mark]]<=r_min
        # value_list = randIntList(lower,upper,len(l_idx[l_mark][res_idxs]))
        for pos in l_idx[l_mark][res_idxs]:
            value = np.random.randint(lower, upper);
            mark[xmin+x_idxs[pos], ymin+y_idxs[pos], zmin+z_idxs[pos]] = value
        # mark[xmin+x_idxs[l_idx[l_mark][res_idxs]], ymin+y_idxs[l_idx[l_mark][res_idxs]], zmin+z_idxs[l_idx[l_mark][res_idxs]]] = 255
        l_mark[l_idx[l_mark][res_idxs]]=False

        # 计算剩余
        if r_max>r_min:
            res_idxs = ((r_max-revert_radius_list[l_idx[l_mark]])*height/(r_max-r_min)) >= revert_coor_mat[l_idx[l_mark],2]
            # value_list = randIntList(lower,upper,len(l_idx[l_mark][res_idxs]))
            for pos in l_idx[l_mark][res_idxs]:
                value = np.random.randint(lower, upper);
                mark[xmin+x_idxs[pos], ymin+y_idxs[pos], zmin+z_idxs[pos]] = value
            # mark[xmin+x_idxs[l_idx[l_mark][res_idxs]], ymin+y_idxs[l_idx[l_mark][res_idxs]], zmin+z_idxs[l_idx[l_mark][res_idxs]]] = 255
            l_mark[l_idx[l_mark][res_idxs]]=False
    else:
        # value_list = randIntList(lower,upper,len(x_idxs))
        for (px,py,pz) in zip(x_idxs,y_idxs,z_idxs):
            value = np.random.randint(lower, upper);
            mark[xmin+x_idxs[px], ymin+y_idxs[py], zmin+z_idxs[pz]] = value
        # mark[xmin+x_idxs, ymin+y_idxs, zmin+z_idxs] = 255



def simulate3DTreeModel(size,max_depth,base_length,base_x,blur=True,level=1):
    MAX_BOX_WIDTH=size
    MAX_TREE_DEPTH = max_depth
    BASE_LENGTH = base_length
    BASE_X=base_x

    # Init space
    mark_shape = (MAX_BOX_WIDTH+1,MAX_BOX_WIDTH+1,MAX_BOX_WIDTH+1)
    mark = np.zeros(mark_shape, dtype=np.uint8)

    # Create root node
    node_count = 0
    root_r = getChildRadius(BASE_X,0,MAX_TREE_DEPTH)
    root_pos = (MAX_BOX_WIDTH//2,MAX_BOX_WIDTH//2,MAX_BOX_WIDTH//2)
    node_count += 1
    root_node = Vertex(node_count,0,root_pos[0],root_pos[1],root_pos[2],root_r,-1)
    setMarkWithSphere(mark, Sphere(Point3D(*root_node.pos), root_node.r), mark_shape, 100, 256)

    # Creante dequeue and list to contain result
    dq = deque([(root_node, 0)]) # 第二项表示node节点的depth
    nodes = {}
    graph = {}

    while len(dq):
        root_node = dq[0][0]
        root_depth = dq[0][1]
        dq.popleft()

        # Add to nodes and graph
        v1 = root_node.idx
        v2 = root_node.p_idx
        if root_node.idx not in nodes:
            nodes[root_node.idx] = root_node
        if v1>0 and v2>0:
            if not v1 in graph:
                graph[v1] = set([v2])
            else:
                graph[v1].add(v2)

            if not v2 in graph:
                graph[v2] = set([v1])
            else:
                graph[v2].add(v1)

        if root_depth<MAX_TREE_DEPTH:
            # Get children number
            if root_node.idx==1: # 根节点单独处理
                child_num = 4
                mask = np.array([[1,1,1],
                                 [-1,1,1],
                                 [1,1,-1],
                                 [-1,1,-1]])
                for i in range(4):
                    # 获取分支半径和长度
                    child_r = getChildRadius(BASE_X,root_depth+1,MAX_TREE_DEPTH)
                    child_length = getChildLength(BASE_LENGTH,root_depth+1,MAX_TREE_DEPTH)

                    #theta_z = np.random.uniform(30,60)
                    theta_y = 45
                    #A = rotation_matrix(theta_z/180*np.math.pi, [0,0,1])
                    B = rotation_matrix(-theta_y/180*np.math.pi, [0,1,0])
                    # rot_mat = np.matmul(A,B)
                    p0 = np.array([[child_length],[0],[0],[1]])
                    p1 = np.matmul(B, p0)
                    child_pos = (int(p1[0]*mask[i][0]+root_node.pos[0]), \
                                 int(p1[1]*mask[i][1]+root_node.pos[1]), \
                                 int(p1[2]*mask[i][2]+root_node.pos[2]))
                    if ImageUtils.bboxCheck3D(child_pos[0], child_pos[1], child_pos[2], child_r, mark_shape):
                        node_count += 1
                        child_node = Vertex(node_count, 0, child_pos[0], child_pos[1], child_pos[2], child_r, root_node.idx)
                        # 绘制
                        setMarkWithSphere(mark, Sphere(Point3D(*child_node.pos), child_node.r), mark_shape, 100, 256)
                        setMarkWithCone(mark, Cone(Point3D(*root_node.pos), root_node.r, \
                                                   Point3D(*child_node.pos), child_node.r), mark_shape, 100, 256)

                        # Add to dequeue
                        dq.append((child_node, root_depth+1))
            else:

                child_num = getRandChildNumber()
                child_angles_range = Draw3DTools.sliceRange(0, 360, child_num)

                for i in range(child_num):

                    # 获取分支半径和长度
                    child_r = getChildRadius(BASE_X,root_depth+1,MAX_TREE_DEPTH)
                    child_length = getChildLength(BASE_LENGTH,root_depth+1,MAX_TREE_DEPTH)

                    # 获取生长角度
                    if child_num==1:
                        theta_z = np.random.uniform(0,360)
                        theta_y = np.random.uniform(60,90)
                    else:
                        theta_z = np.random.uniform(child_angles_range[i][0],child_angles_range[i][1])
                        theta_y = np.random.uniform(30,70)

                    A = rotation_matrix(theta_z/180*np.math.pi, [0,0,1])
                    B = rotation_matrix(-theta_y/180*np.math.pi, [0,1,0])
                    rot_mat = np.matmul(A,B)
                    p0 = np.array([[child_length],[0],[0],[1]])
                    p1 = np.matmul(rot_mat, p0)

                    grand_node = nodes[root_node.p_idx] # root节点的父节点
                    p_a = Point3D(0,0,0)
                    p_c = Point3D(root_node.pos[0]-grand_node.pos[0], \
                                  root_node.pos[1]-grand_node.pos[1], \
                                  root_node.pos[2]-grand_node.pos[2])
                    p_b = p_a.medianWithPoint(p_c)
                    v1=np.array([[p_a.x, p_b.x, p_c.x], #局部坐标点
                                 [p_a.y, p_b.y, p_c.y],
                                 [p_a.z, p_b.z, p_c.z],
                                 [   1,    1,    1]])
                    Dis=p_a.distanceWithPoint(p_c)
                    v0=np.array([[0,     0,   0], # 局部坐标点
                                 [0,     0,   0],
                                 [-Dis, -Dis/2, 0],
                                 [1,     1,  1]])
                    rev_mat = superimposition_matrix(v0,v1)
                    p2 = np.matmul(rev_mat, p1)
                    child_pos = (int(p2[0]+grand_node.pos[0]), int(p2[1]+grand_node.pos[1]), int(p2[2]+grand_node.pos[2]))
                    if ImageUtils.bboxCheck3D(child_pos[0], child_pos[1], child_pos[2], child_r, mark_shape):
                        node_count += 1
                        child_node = Vertex(node_count, 0, child_pos[0], child_pos[1], child_pos[2], child_r, root_node.idx)
                        # 绘制
                        setMarkWithSphere(mark, Sphere(Point3D(*child_node.pos), child_node.r), mark_shape, 100, 256)
                        setMarkWithCone(mark, Cone(Point3D(*root_node.pos), root_node.r, \
                                                   Point3D(*child_node.pos), child_node.r), mark_shape, 100, 256)

                        # Add to dequeue
                        dq.append((child_node, root_depth+1))


    mark = mark.astype(np.uint8)
    mark = np.swapaxes(mark, 0, 2)
    if blur:
        mark=Draw3DTools.gaussFilter3DVolume(mark,6*level,2*level,2*level)

    # 标准化到0-255
    mark = ImageUtils.normalizeImage(mark)

    return mark, nodes, graph



# @Time    : 2021/8/29
# @Author  : zhanghan

import os
import copy
import time
import matplotlib.pyplot as plt
import pandas as pd
import importlib

from pyneval.tools.optimize.SA import SAFast
from pyneval.metric import ssd_metric
from pyneval.metric.utils import config_utils
from pyneval.model import swc_node
from pyneval.io import read_json
from pyneval.metric.utils import metric_manager


def SA_optimize(gold_tree=None, metric_method=None, config=None, 
                metric_config=None, optimize_config=None,
                test_name=None, lock=None):
    # fullfill cmd 
    CURRENT_TEST_PATH = os.path.join(optimize_config["trace"]["workDir"], test_name+"_test.swc")
    REC_CMD = optimize_config["trace"]["cli"].format(CURRENT_TEST_PATH, config)
    # run cmd
    try:
        os.system(REC_CMD)
    except:
        raise Exception("[Error: ] error executing reconstruction")
    # run metric
    res_tree = swc_node.SwcTree()
    res_tree.load(CURRENT_TEST_PATH)

    if lock is not None:
        lock.acquire()
    try:
        main_score, _, _ = metric_method(gold_tree, res_tree, metric_config)
    finally:
        if lock is not None:
            lock.release()
    # avg
    # score = (main_score["recall"] + main_score["precision"])/2
    # f1
    try:
        score = 2*main_score["recall"]*main_score["precision"]/(main_score["recall"] + main_score["precision"])
    except:
        score = 0
    print("[Info: ] ssd loss = {}".format(score))

    return config, -score



def check_save(gold_tree, metric_method, metric_config, optimize_config, best_name, best_configs):
    # fullfill cmd 
    BEST_TEST_PATH = os.path.join(optimize_config["trace"]["workDir"], best_name+"_best.swc")
    REC_CMD = optimize_config["trace"]["cli"].format(BEST_TEST_PATH, best_configs)
    # run cmd
    try:
        os.system(REC_CMD)
    except:
        raise Exception("[Error: ] error executing reconstruction")
    # run metric
    res_tree = swc_node.SwcTree()
    res_tree.load(BEST_TEST_PATH)

    main_score, _, _ = metric_method(gold_tree, res_tree, metric_config)
    try:
        score = 2*main_score["recall"]*main_score["precision"]/(main_score["recall"] + main_score["precision"])
    except:
        score = 0
    for key in main_score:
        print("{}:{}".format(key, main_score[key]))
    print("f1 score:{}".format(score))

def optimize(gold_swc_tree, test_swc_paths, optimize_config, metric_config, metric_method):
    start_timestep = time.time()
    # set test tiff path
    io_dict = {
        "input": test_swc_paths[0],
        "output": "{0}"
    }
    
    # put normal parameters into cmd, except for output swc
    optimize_config["trace"]["cli"] = optimize_config["trace"]["cli"].format(**io_dict)
    # add target parameters into cmd 
    it = 0
    for key in optimize_config["trace"]["parameters"]:
        optimize_config["trace"]["cli"] += (" " + key + " {1[" + str(it) + "]}")
        it+=1
    # initialize and run SA model
    sa_fast = SAFast(func=SA_optimize, gold_swc_tree=gold_swc_tree, metric_method=metric_method, 
                    metric_config=metric_config, optimize_config=optimize_config)
    best_configs, best_value = sa_fast.run(gold_swc_tree=gold_swc_tree, metric_method=metric_method, 
                                        metric_config=metric_config, optimize_config=optimize_config)
    # output the best parameters
    print(best_configs)
    i=0
    print("[Info: ]best configs:")
    for key in optimize_config["trace"]["parameters"]:
        print("best {} = {}".format(key, best_configs[i]))
        i+=1
    print(
        "best value = {}\n"
        "time = {}\n" .format(-best_value, time.time() - start_timestep)
    )
    # get and save best reconstruct swc
    print("[Info: ]Exam test score with best configs: ")
    best_name = test_swc_paths[0][:-4]
    check_save(gold_tree=gold_swc_tree, metric_method=metric_method, 
               metric_config=metric_config, optimize_config=optimize_config, 
               best_name=best_name, best_configs=best_configs)

    # plot the result.
    plt.plot(pd.DataFrame(sa_fast.best_y_history).cummin(axis=0))
    plt.xlabel("iterations")
    plt.ylabel("score(opposite value)")
    print("[Info:]Figure saved in {}".format(os.path.join(optimize_config["trace"]["workDir"], best_name+".png")))
    plt.savefig(os.path.join(optimize_config["trace"]["workDir"], best_name+".png"))
    plt.show()
    return 0



import math
import copy
import queue
from pyneval.model.swc_node import SwcNode,SwcTree
from pyneval.model.binary_node import BinaryNode
from pyneval.model.euclidean_point import EuclideanPoint
from pyneval.metric.utils.bin_utils import calculate_trajectories_xy, calculate_trajectories_z

_2D = "2d"

TRAJECTORY_NONE = -1.0
global g_xy_threshold
global g_z_threshold
global g_default_xy_path_error_threshold
global g_default_z_path_error_threshold
global g_local_path_error_threshold


def diadem_utils_init(config):
    global TRAJECTORY_NONE
    global g_xy_threshold
    global g_z_threshold
    global g_default_xy_path_error_threshold
    global g_default_z_path_error_threshold
    global g_local_path_error_threshold

    g_xy_threshold = config["xy_threshold"]
    g_z_threshold = config["z_threshold"]
    g_default_xy_path_error_threshold = config["default_xy_path_error_threshold"]
    g_default_z_path_error_threshold = config["default_z_path_error_threshold"]
    g_local_path_error_threshold = config["local_path_error_threshold"]


def to_swc_node(node_a, node_b):
    if isinstance(node_a, BinaryNode):
        node_a = node_a.data
    if isinstance(node_b, BinaryNode):
        node_b = node_b.data
    return node_a, node_b


def get_end_node_XY_dis_diff(gold_swc_node,
                             trajectory,
                             test_swc_node):
    gold_dist = gold_swc_node.distance(trajectory, mode=_2D)
    test_dist = test_swc_node.distance(trajectory, mode=_2D)
    return gold_dist - test_dist


def get_end_node_Z_dis_diff(gold_swc_node,
                            trajectory,
                            test_swc_node):
    gold_dist = math.fabs(gold_swc_node.get_z() - trajectory.get_z())
    test_dist = math.fabs(test_swc_node.get_z() - trajectory.get_z())
    return gold_dist - test_dist


def get_nearby_node_list(gold_node=None, bin_test_list=None, check_previous_use=True, g_matches={}):
    nearby_node_list = []

    for test_node in bin_test_list:
        if (not check_previous_use or test_node not in g_matches.keys()) and is_in_threshold(gold_node, test_node):
            nearby_node_list.append(test_node)

    return nearby_node_list


def is_in_XY_threshold(gold_node, test_node):
    gold_node, test_node = to_swc_node(gold_node, test_node)

    xy_diff = math.sqrt((gold_node.get_x() - test_node.get_x())**2 + (gold_node.get_y() - test_node.get_y())**2)

    if xy_diff <= g_xy_threshold:
        return True
    else:
        return False


def is_in_threshold(gold_node, test_node):
    gold_node, test_node = to_swc_node(gold_node, test_node)

    xy_diff = math.sqrt((gold_node.get_x() - test_node.get_x())**2 + (gold_node.get_y() - test_node.get_y())**2)
    z_diff = math.fabs(gold_node.get_z() - test_node.get_z())

    if xy_diff <= g_xy_threshold and z_diff <= g_z_threshold + 0.1:
        return True
    else:
        return False


def get_trajectory_for_path(ancestor_node, descendant_node):
    path_stack = queue.LifoQueue()
    ancestor_data = ancestor_node.data

    while descendant_node != ancestor_node:
        path_stack.put(descendant_node)
        descendant_node = descendant_node.parent
        if descendant_node is None:
            return None

    trajectory = EuclideanPoint()

    descendant_node = path_stack.get()
    down_x = False
    down_z = False

    while not path_stack.empty() and (not down_x or not down_z):
        sec_data = descendant_node.data
        next_descendant = path_stack.get()

        if next_descendant.is_left():
            if not down_x and sec_data.left_trajectory.get_x() != TRAJECTORY_NONE:
                trajectory.set_x(sec_data.left_trajectory.get_x())
                trajectory.set_y(sec_data.left_trajectory.get_y())
                down_x = True
            if not down_z and sec_data.left_trajectory.get_z() != TRAJECTORY_NONE:
                trajectory.set_z(sec_data.left_trajectory.get_z())
                down_z = True
        elif not next_descendant.is_left() and sec_data.right_trajectory is not None:
            if not down_x and sec_data.right_trajectory.get_x() != TRAJECTORY_NONE:
                trajectory.set_x(sec_data.right_trajectory.get_x())
                trajectory.set_y(sec_data.right_trajectory.get_y())
                down_x = True
            if not down_z and sec_data.right_trajectory.get_z() != TRAJECTORY_NONE:
                trajectory.set_z(sec_data.right_trajectory.get_z())
                down_z = True

        if not down_x and not is_in_XY_threshold(ancestor_data, next_descendant.data):
            tmp = calculate_trajectories_xy(ancestor_data,
                                            descendant_node.data,
                                            next_descendant.data,
                                            g_xy_threshold)
            trajectory.set_x(tmp.get_x())
            down_x = True

        if not down_z and math.fabs(ancestor_data.get_z() - next_descendant.get_z()) > g_z_threshold:
            trajectory.set_z(calculate_trajectories_z(ancestor_data,
                                                      descendant_node.data,
                                                      next_descendant.data,
                                                      g_z_threshold))
            down_z = True
        descendant_node = next_descendant

    if not down_x:
        trajectory.set_x(descendant_node.data.get_x())
        trajectory.set_y(descendant_node.data.get_y())

    if not down_z:
        trajectory.set_z(descendant_node.data.get_z())

    return trajectory


def path_length_matches(gold_swc_path_length,
                        test_XY_path_length,
                        test_Z_path_length,
                        DEBUG = False):
    xy_path_error_threshold = g_default_xy_path_error_threshold
    z_path_error_threshold = g_default_z_path_error_threshold

    xy_diff = math.fabs(gold_swc_path_length.xy_path_length - test_XY_path_length)
    if gold_swc_path_length.path_length == 0:
        return 0
    xy_err = xy_diff / gold_swc_path_length.path_length

    if DEBUG:
        print("[Info:  ]In path length matches xy_diff = {}, xy_err = {}".format(
            xy_diff, xy_err
        ))

    z_diff = math.fabs(gold_swc_path_length.z_path_length - test_Z_path_length)
    z_err = z_diff / gold_swc_path_length.path_length

    if gold_swc_path_length.xy_path_length < g_xy_threshold:
        if test_XY_path_length < g_xy_threshold:
            xy_err = 0
        else:
            xy_path_error_threshold = g_local_path_error_threshold

    if gold_swc_path_length.z_path_length < g_z_threshold:
        if test_Z_path_length < g_z_threshold:
            z_err = 0
        else:
            z_path_error_threshold = g_local_path_error_threshold

    if xy_err < g_default_xy_path_error_threshold and z_err < g_default_z_path_error_threshold:
        return (xy_err/xy_path_error_threshold + z_err/z_path_error_threshold)/2
    else:
        return 1


def get_match_path_length_difference(gold_node, test_node, bin_gold_list, bin_test_list):
    if test_node.parent is None:
        return 0
    gold_target = gold_node

    gold_swc_path_length = SwcNode()
    gold_swc_path_length.add_length(gold_node.data)
    test_swc_path_length = SwcNode()
    test_swc_path_length.add_length(test_node.data)

    test_path_XY_mod = get_end_node_XY_dis_diff(
        gold_node.data, gold_node.data.parent_trajectory, test_node.data
    )
    test_path_Z_mod = get_end_node_Z_dis_diff(
        gold_node.data, gold_node.data.parent_trajectory, test_node.data
    )
    test_swc_path_length.xy_path_length += test_path_XY_mod
    test_swc_path_length.z_path_length += test_path_Z_mod

    is_branch_left = gold_node.is_left()
    gold_node = gold_node.parent
    test_node = test_node.parent

    ancestor_trajectory = EuclideanPoint()
    checked_node = set()
    no_match = True
    no_done = True

    while no_match and no_done:
        if is_in_threshold(gold_node, test_node):
            gold_data = gold_node.data

            if is_branch_left:
                ancestor_trajectory = gold_data.left_trajectory
            else:
                ancestor_trajectory = gold_data.right_trajectory

            if ancestor_trajectory.get_x() == TRAJECTORY_NONE or \
                ancestor_trajectory.get_z() == TRAJECTORY_NONE:
                tmp_trajectory = get_trajectory_for_path(gold_node, gold_target)
                if ancestor_trajectory.get_x() == TRAJECTORY_NONE:
                    ancestor_trajectory.set_x(tmp_trajectory.get_x())
                    ancestor_trajectory.set_y(tmp_trajectory.get_y())
                if ancestor_trajectory.get_z() == TRAJECTORY_NONE:
                    ancestor_trajectory.set_z(tmp_trajectory.get_z())

            test_XY_path_length = test_swc_path_length.xy_path_length + \
                                  get_end_node_XY_dis_diff(gold_data, ancestor_trajectory, test_node.data)
            test_Z_path_length = test_swc_path_length.z_path_length + \
                                 get_end_node_Z_dis_diff(gold_data, ancestor_trajectory, test_node.data)
            percent_error = path_length_matches(gold_swc_path_length, \
                                               test_XY_path_length, \
                                               test_Z_path_length)
            if percent_error < 1:
                return percent_error
            else:
                return 1
        else:
            if gold_swc_path_length.path_length < test_swc_path_length.path_length:
                if gold_node.parent is None:
                    no_done = False
                else:
                    nearby_nodes = get_nearby_node_list(gold_node, bin_test_list, False)
                    for node in nearby_nodes:
                        if node in checked_node:
                            no_done = False
                            break
                    checked_node.add(gold_node)
                    gold_swc_path_length.add_length(gold_node.data)
                    is_branch_left = gold_node.is_left()
                    gold_node = gold_node.parent
            else:
                if test_node.parent is None:
                    no_done = False
                else:
                    nearby_nodes = get_nearby_node_list(test_node, bin_gold_list, False)
                    for node in nearby_nodes:
                        if node in checked_node:
                            no_done = False
                            break

                    checked_node.add(test_node)
                    test_swc_path_length.add_length(test_node.data)
                    test_node = test_node.parent
    return 1


def LCA(node1, node2, kca):
    node1_list = []

    node1 = node1.parent
    while node1 != kca:
        node1_list.append(node1)
        node1 = node1.parent

    node1 = node2.parent
    while node2 != kca and node2 is not None:
        if node2 in node1_list:
            return node2
        node2 = node2.parent

    return None


def is_within_dis_match_threshold(node1, node2):
    if isinstance(node1, BinaryNode):
        node1 = node1.data
    if isinstance(node2, BinaryNode):
        node2 = node2.data
    return node1.distance(node2, _2D) <= g_xy_threshold * 3 \
           and math.fabs(node1.get_z() - node2.get_z()) < g_z_threshold * 3 + 0.1



"""Draw2DTools, providing functions to generate 2d images
"""
from functools import reduce
import numpy as np
from scipy.ndimage import filters as ndfilter
import matplotlib.pyplot as plt
import matplotlib.lines as lines
from mpl_toolkits.mplot3d.art3d import Line3D
import matplotlib.transforms as mtransforms
import matplotlib.text as mtext

from .Utils import joinPath
from .PlotTools import createNewSubplot
from .ImageUtils import ImgType, OpType

def randIntList(lower, upper, length):
    random_list = []
    for i in range(length):
        random_list.append(np.random.randint(lower, upper))
    return random_list



def gaussFilter3DVolume(im,sz,sx,sy):
    # s0 = np.random.uniform(1.0,2.0)
    # s1 = np.random.uniform(0.5,1.0)
    # s2 = np.random.uniform(0.5,1.0)

    im = ndfilter.gaussian_filter(im, [sz,sx,sy])
    return im

class MyLine(lines.Line2D):
    def __init__(self, *args, **kwargs):
        # we'll update the position when the line data is set
        self.text = mtext.Text(0, 0, '')
        lines.Line2D.__init__(self, *args, **kwargs)

        # we can't access the label attr until *after* the line is
        # inited
        self.text.set_text(self.get_label())

    def set_figure(self, figure):
        self.text.set_figure(figure)
        lines.Line2D.set_figure(self, figure)

    def set_axes(self, axes):
        self.text.set_axes(axes)
        lines.Line2D.set_axes(self, axes)

    def set_transform(self, transform):
        # 2 pixel offset
        texttrans = transform + mtransforms.Affine2D().translate(2, 2)
        self.text.set_transform(texttrans)
        lines.Line2D.set_transform(self, transform)

    def set_data(self, x, y):
        if len(x):
            self.text.set_position((x[-1], y[-1]))

        lines.Line2D.set_data(self, x, y)

    def draw(self, renderer):
        # draw my label at the end of the line with 2 pixel offset
        lines.Line2D.draw(self, renderer)
        self.text.draw(renderer)

def getVertexSequence(pa,pb, line_width_range):
    ''' Separate va-vb to more segments, and returen the points list as well as line_widths list
        Args:
            pa (tuple): start point
            pb (tuple): end point
            line_width_range (tuple): line width range
    '''
    segs = np.random.randint(20,30)

    pts = [pa]
    for i in range(1,segs):
        pc = tuple([pa[j]+(pb[j]-pa[j])*i*1.0/segs for j in range(3)])
        pts.append(pc)
    pts.append(pb)

    line_widths = []

    line_segs = [([pts[idx][0], pts[idx+1][0]],[pts[idx][1], pts[idx+1][1]],[pts[idx][2], pts[idx+1][2]]) for idx in range(segs)]
    line_widths = np.linspace(line_width_range[0], line_width_range[1], segs)
#     print(line_segs)
#     print('line width: ',line_widths)
    return line_segs, line_widths

def drawGradLines(ax, line_segs, line_widths):
    for l,w in zip(line_segs, line_widths):
        ax.add_line(Line3D(l[0], l[1], l[2], lw=w, c='k'))

def sliceRange(s, t, n):
    assert(t>=1)
    if n>1:
        arr, step = np.linspace(s, t, n, endpoint=False, retstep=True)
        arr = [(x+step/3, x+step*2/3) for x in arr]
    else: # n==1
        step = t - s
        arr = [(s+step/3,s+step*2/3)]
    return arr

def rotMatrix3d(theta, axis):
    '''rotMatrix3d: Given axis and theta, calculate rotation matrix in 3d axes.
    Args:
        theta (int): The negative angle along  right-hand screw rule.
        axis (string): The axis which rotate along with.
            'x','y' or 'z'
    Returns:
        M (numpy array): The rotation matrix.
    '''
    assert axis in ['x','y','z']
    c, s = np.cos(theta/180*np.pi), np.sin(theta/180*np.pi)
    if axis=='x':
        M = np.matrix([[1, 0, 0, 0],
                       [0, c, -s,0],
                       [0, s, c, 0],
                       [0, 0, 0, 1]])
    elif axis=='y':
        M = np.matrix([[c, 0, s, 0],
                       [0, 1, 0,  0],
                       [-s,0, c,  0],
                       [0, 0, 0, 1]])
    else: # 'z'
        M = np.matrix([[c, -s, 0, 0],
                       [s, c,  0,0],
                       [0, 0, 1, 0],
                       [0, 0, 0, 1]])
    return M

def transMatrix3d(tx, ty, tz):
    M = np.matrix([[1,0,0,tx],
                   [0,1,0,ty],
                   [0,0,1,tz],
                   [0,0,0,1]])
    return M

def transMatrix3dCompose(theta_z, theta_y, tx, ty, tz):
    m_r_z = rotMatrix3d(theta_z, 'z')
    m_r_y = rotMatrix3d(theta_y, 'y')
    m_t = transMatrix3d(tx, ty, tz)
    return reduce(np.matmul, [m_t, m_r_y, m_r_z])

def appendMatrix3d(m, m_add):
    return np.matmul(m, m_add)

def cosAngle(x):
    return np.cos(x/180*np.pi)

def sinAngle(x):
    return np.sin(x/180*np.pi)

def getRandomTransformParas(theta_z_range,
                            theta_y_range,
                            r_range):
    theta_z = np.random.uniform(theta_z_range[0],theta_z_range[1])
    theta_y = np.random.uniform(theta_y_range[0],theta_y_range[1])
    r = np.random.randint(r_range[0], r_range[1])
    return theta_z, theta_y, r

def grow3(M, ps, theta_z, theta_y, r, ax, line_width_range):
    """ grow3, 在局部坐标系第一象限绘制
    """
    # Calculate local coordination
    p_l = np.array([[r],[0],[0],[1]])
    m_rot_z = rotMatrix3d(theta_z, 'z')
    m_rot_y = rotMatrix3d(-theta_y, 'y')
    m_rot_all = np.matmul(m_rot_y, m_rot_z)
    p_l = np.matmul(m_rot_all, p_l)

    # Transform to global coordination
    p = np.matmul(M, p_l)
    p = tuple(np.asarray(p).flatten())

    # Draw line segments
    line_segs, line_widths = getVertexSequence(ps, p, line_width_range)
    drawGradLines(ax, line_segs, line_widths)

    M_l = transMatrix3dCompose(theta_z, theta_y, p_l[0], p_l[1], p_l[2])
    M = appendMatrix3d(M, M_l)
    return M, p

def expand(revert_matrix, start_point, ax, depth, img_type, max_width):
    if depth>0:
        # Rotate along z clockwise 90
        revert_matrix = appendMatrix3d(revert_matrix, rotMatrix3d(90, 'y'))

        if img_type==ImgType.Branch:
            branch_numbers = np.random.randint(2,4)
            angle_slices = sliceRange(0, 360, branch_numbers)
            for angle in angle_slices:
                angle_s = angle[0]
                angle_width = angle[1]-angle[0]

                revert_matrix_local = appendMatrix3d(revert_matrix, rotMatrix3d(angle_s, 'z'))
                theta_range = (0, angle_width)
                radiu_range = (3*depth+1, 3*depth+9)
                local_max_width = max_width
                local_min_width = max_width-np.random.randint(1, 11)
                line_width_range = (local_max_width, local_min_width)
                # line_width_range = (5*depth+15, np.random.randint(5*depth+10, 5*depth+16))
                theta_z, theta_y, radiu = getRandomTransformParas(theta_range, (75,85), radiu_range)
                revert_matrix_new, start_point_new = grow3(revert_matrix_local,
                                                           start_point,
                                                           theta_z,
                                                           theta_y,
                                                           radiu,
                                                           ax,
                                                           line_width_range)
                expand(revert_matrix_new, start_point_new, ax, depth-1, img_type, local_min_width)

        else: # NonBranch
            angle_s = 0
            angle_width = 360

            revert_matrix_local = appendMatrix3d(revert_matrix, rotMatrix3d(angle_s, 'z'))
            theta_range = (0, angle_width)
            radiu_range = (8, 14)
            line_width_range = (max_width, np.random.randint(max_width-5,max_width+1))
            theta_z, theta_y, radiu = getRandomTransformParas(theta_range, (75, 85), radiu_range)
            revert_matrix_new, start_point_new = grow3(revert_matrix_local,
                                                       start_point,
                                                       theta_z,
                                                       theta_y,
                                                       radiu,
                                                       ax,
                                                       line_width_range)

def initCanvas3D():
    ax = createNewSubplot(111, projection='3d')

    fig = ax.get_figure()
    fig.set_size_inches(10, 10)
    ax.set_xlim(-10,10)
    ax.set_ylim(-10,10)
    ax.set_zlim(-10,10)
    # ax.set_xlim(-10,10)
    # ax.set_ylim(-10,10)

    return fig,ax


def drawWithGrowing(ax, img_type, dest_path, img_name):
    ax.cla()



    if img_type==ImgType.Branch:
        start_point = (-10, -10,-10, 1)
        revert_matrix = transMatrix3d(-10,-10, -10)

        max_width = np.random.randint(30, 51)
        min_width = max_width - np.random.randint(1, 11)
        line_width_range = (max_width, min_width)
        radiu_range = (15,17)
        theta_z_range = (30,60)
        theta_y_range = (30,60)
        theta_z, theta_y,radiu = getRandomTransformParas(theta_z_range, theta_y_range, radiu_range)
        revert_matrix, start_point = grow3(revert_matrix,
                                                  start_point,
                                                  theta_z,
                                                  theta_y,
                                                  radiu,
                                                  ax,
                                                  line_width_range)

        expand_depth = 1 + (np.random.uniform()>0.8);
        expand(revert_matrix, start_point, ax, expand_depth, img_type, min_width)
    else:
        start_point = (-10, -10, -10, 1)
        revert_matrix = transMatrix3d(-10,-10,-10)

        max_width = np.random.randint(30, 51)
        min_width = np.random.randint(max_width-5,max_width+1)
        line_width_range = (max_width, min_width)
        radiu_range = (20,23)
        theta_z_range = (30,60)
        theta_y_range = (30,60)
        theta_z, theta_y,radiu = getRandomTransformParas(theta_z_range, theta_y_range, radiu_range)
        revert_matrix, start_point = grow3(revert_matrix,
                                           start_point,
                                           theta_z,
                                           theta_y,
                                           radiu,
                                           ax,
                                           line_width_range)

        expand(revert_matrix, start_point, ax, 1, img_type, min_width)

    # plt.axis('off')
    plt.tight_layout(pad=0)
    plt.savefig(joinPath(dest_path, img_name), pad_inches=0, frameon=False, dpi=4)
    # plt.savefig(joinPath(dest_path, img_name), bbox_inches='tight', pad_inches=0, frameon=False)

def generateLines(img_type):
    if img_type==ImgType.NonBranch:
        neighbor_number = np.random.randint(1,3)  # 1~2

    else:
        neighbor_number = np.random.randint(3,5) # 3~4

    lines_array = []
    center_r = np.random.randint(0, 4) # 1~5
    center_angle = np.random.randint(0, 360)
    x0 = int(center_r * np.cos(np.pi * center_angle/180))
    y0 = int(center_r * np.sin(np.pi * center_angle/180))

    for i in range(neighbor_number):
        neighbor_r = np.random.randint(6, 11) # 6~10
        neighbor_angle = np.random.randint(0, 360)
        x1 = int(neighbor_r * np.cos(np.pi * neighbor_angle/180))
        y1 = int(neighbor_r * np.sin(np.pi * neighbor_angle/180))

        # print x0, y0, x1, y1, neighbor_angle
        lines_array.append(MyLine([x0, x1], [y0, y1], lw=20, c='k'))

    return lines_array



def drawImage(img_type, dest_path, img_name):
#     fig = plt.figure()
    plt.clf()
    fig = plt.gcf()
    fig.set_size_inches(10, 10)
    ax = plt.gca(xlim=[-10,10],ylim=[-10,10]) # Set x and y range

    lines_array = generateLines(img_type)
    for l in lines_array:
        ax.add_line(l)

    # plt.axis('off')
    plt.tight_layout(pad=0)
    plt.savefig(joinPath(dest_path,img_name), bbox_inches='tight', pad_inches=0, frameon=False, dpi=4)



import os

from pyneval.errors.exceptions import InvalidSwcFileError
from pyneval.model.swc_node import SwcNode, SwcTree


def is_swc_file(file_path):
    return file_path[-4:] in (".swc", ".SWC")

def read_swc_tree(swc_file_path):
    if not os.path.isfile(swc_file_path) or not is_swc_file(swc_file_path):
        raise InvalidSwcFileError(swc_file_path)
    swc_tree = SwcTree()
    swc_tree.load(swc_file_path)
    return swc_tree

# if path is a folder
def read_swc_trees(swc_file_paths, tree_name_dict=None):
    """
    Read a swc tree or recursively read all the swc trees in a fold
    Args:
        swc_file_paths(string): path to read swc
        tree_name_dict(dict): a map for swc tree and its file name
            key(SwcTree): SwcTree object
            value(string): name of the swc tree
    Output:
        swc_tree_list(list): a list shaped 1*n, containing all the swc tree in the path
    """
    # get all swc files
    swc_files = []
    if os.path.isfile(swc_file_paths):
        if not is_swc_file(swc_file_paths):
            print(swc_file_paths + "is not a swc file")
            return
        swc_files = [swc_file_paths]
    else:
        for root, _, files in os.walk(swc_file_paths):
            for file in files:
                f = os.path.join(root, file)
                if is_swc_file(f):
                    swc_files.append(f)
    # load swc trees
    swc_tree_list = []
    for swc_file in swc_files:
        swc_tree = SwcTree()
        swc_tree.load(swc_file)
        swc_tree_list.append(swc_tree)
        if tree_name_dict is not None:
            tree_name_dict[swc_tree] = os.path.basename(swc_file)
    return swc_tree_list


def adjust_swcfile(swc_str):
    words = swc_str.split("\n")
    return words


def read_from_str(swc_str):
    swc_tree = SwcTree()
    swc_list = adjust_swcfile(swc_str)
    swc_tree.load_list(swc_list)
    return swc_tree



# bennieHan 2019-11-12 16:01
# all right reserved

import queue
import math
import time
import queue
import copy
from pyneval.metric.utils.bin_utils import convert_to_binarytree,swctree_to_binarytree
from pyneval.model.swc_node import SwcNode, SwcTree


# testfunction:
# input: a swc tree root
# return: none
# print the whole tree
def test_print_tree(root_node):
    num = 0
    q = queue.Queue()
    q.put(root_node)
    while not q.empty():
        num += 1
        cur = q.get()
        for item in cur.children:
            q.put(item)
        print("-----------")
    print("tree node num = {}".format(num))

def test_print_bin_tree(root_node):
    num = 0
    q = queue.Queue()
    q.put(root_node)
    while not q.empty():
        # num += 1
        cur = q.get()
        # if ((cur.left_son != None and cur.right_son != None) or \
        #        (cur.left_son == None and cur.right_son == None)):
        #     # print("node_id = {}, Left = {}, Right = {}".format(cur.data.id,cur.left_son, cur.right_son))
        #     # print("出大问题！")
        #     num += 1
        if cur.left_son != None:
            # print(cur.left_son.data.id)
            q.put(cur.left_son)
        if cur.right_son != None:
            # print(cur.right_son.data.id)
            q.put(cur.right_son)
        # print("-----------")
    print("tree node num = {}".format(num))

#　测试说明：
# set的查找比list快很多,set和list黄转换挺快的，set到list更快一点
def test1():
    a = set()
    start = time.time()
    for i in range(10000000):
        a.add(i)
    print(time.time() - start)
    a = list(a)
    print(time.time() - start)
    a = set(a)
    print(time.time() - start)
    num = 0
    for i in range(5000000,15000000):
        if i in a:
            num += 1;
    print(num)
    print(time.time() - start)

# 测试一下读入整棵树的时间
def test2():
    start = time.time()
    swc_tree = SwcTree()
    swc_tree.load("D:\gitProject\mine\PyNeval\\test\data_example\gold\ExampleGoldStandard.swc")
    bin_tree = convert_to_binarytree(swc_tree)
    test_print_bin_tree(bin_tree)
    print(time.time() - start)

if __name__ == "__main__":
    test2()


import argparse
import importlib
import os
import sys
import jsonschema
import pkg_resources

from multiprocessing import Pool, cpu_count
from pyneval.errors.exceptions import InvalidMetricError, PyNevalError
from pyneval.io import read_json
from pyneval.io.read_swc import read_swc_tree, read_swc_trees
from pyneval.io.swc_writer import swc_save
from pyneval.metric.utils import anno_utils, config_utils
from pyneval.metric.utils.metric_manager import get_metric_manager
from pyneval.tools.optimize import optimize

# load method in metrics
def import_metrics():
    base_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    metric_path = os.path.join(base_dir, "pyneval/metric")
    files = os.listdir(metric_path)
    metrics = []
    for f in files:
        m_f = f.split(".")
        if len(m_f) == 2 and m_f[0][-7:] == "_metric" and m_f[1] == "py":
            metrics.append(m_f[0])
    for m in metrics:
        md = "pyneval.metric.{}".format(m)
        importlib.import_module(md)


def read_parameters():
    metric_manager = get_metric_manager()

    parser = argparse.ArgumentParser(description="Current version: pyneval {}".format(
        pkg_resources.require("pyneval")[0].version)
    )

    parser.add_argument(
        "--gold",
        "-G",
        help="path of the gold standard SWC file",
        required=False
    )
    parser.add_argument(
        "--test",
        "-T",
        help="a list of reconstructed SWC files or folders for evaluation",
        required=False,
        nargs="*"
    )
    parser.add_argument(
        "--metric",
        "-M",
        help="metric choice: " + metric_manager.get_metric_summary(False) + ".",
        required=False
    )
    parser.add_argument(
        "--output",
        "-O",
        help="output path of metric results, output file is in json format with different scores of the metric",
        required=False,
    )
    parser.add_argument(
        "--detail",
        "-D",
        help="output path of detail metric result, swc format presented.\n"
             "identify different type according to metric result for each node",
        required=False,
    )
    parser.add_argument(
        "--config",
        "-C",
        help="path of custom configuration file for the specified metric",
        required=False,
    )
    parser.add_argument(
        "--parallel",
        "-P",
        help="Enable the parallel processing",
        required=False,
        action="store_true"
    )
    parser.add_argument(
        "--optimize",
        help="Enable optimizer mode",
        required=False,
    )

    parser.add_argument("--debug", help="print debug info or not", required=False, action="store_true")

    return parser.parse_args()


def init(abs_dir):
    sys.path.append(abs_dir)
    sys.path.append(os.path.join(abs_dir, "src"))
    sys.path.append(os.path.join(abs_dir, "test"))
    sys.setrecursionlimit(1000000)


# def check_dir(dir_name, dir):
#     while not os.path.exists(dir) or not os.path.isdir(dir):
#         print(
#             "The input path {} for {} does not exist or is not a folder. You may choose to:\n"
#             "[Input=1]Input a new path\n"
#             "[Input=2]Quit this process\n"
#             "[Input=3]Continue without saving\n"
#             "[Input=4]Create new folder {}.".format(dir, dir_name, dir)
#         )
#         choice = input()
#         if choice.lower() == "1":
#             print("Input new detail path:")
#             new_dirorpath = input()
#             if new_dirorpath[-5:] == ".json":
#                 new_dir = os.path.dirname(new_dirorpath)
#             else:
#                 new_dir = new_dirorpath
#             if os.path.exists(new_dir) and os.path.isdir(new_dir):
#                 return 1, new_dirorpath
#             else:
#                 dir = new_dir
#                 continue
#         elif choice.lower() == "2":
#             print("Pyneval ends...")
#             return 2, ""
#         elif choice.lower() == "3":
#             print("Pyneval processing without saving {}...".format(dir_name))
#             return 3, ""
#         elif choice.lower() == "4":
#             if os.path.isfile(dir):
#                 print("[Info: ] Error input")
#                 continue
#             os.makedirs(dir)
#             print("{} has been created".format(dir))
#             return 4, ""
#         else:
#             print("[Info: ] Error input")
#     return 4, ""


def set_configs(abs_dir, args):
    # argument: debug
    is_debug = False
    if args.debug and args.debug.lower() in ("true", "t", "yes"):
        is_debug = True

    # argument: gold
    gold_swc_path = os.path.join(abs_dir, args.gold)
    gold_swc_tree = read_swc_tree(gold_swc_path)  # SwcTree

    # argument: metric
    metric_manager = get_metric_manager()
    metric = metric_manager.get_root_metric(args.metric)
    if not metric:
        raise InvalidMetricError(args.metric, metric_manager.get_metric_summary(True))

    # argument: test
    test_swc_paths = [os.path.join(abs_dir, path) for path in args.test]
    test_swc_trees = []
    # read test trees
    for file in test_swc_paths:
        if file[-4:].lower() == ".tif":
            continue
        test_swc_trees.extend(read_swc_trees(file))


    if len(test_swc_paths) == 0:
        raise PyNevalError("test models can't be null")

    # info: how many trees read
    print("Evaluating {} test model(s) \n".format(len(test_swc_trees)))

    # argument: config
    config_path = args.config
    if config_path is None:
        config = config_utils.get_default_configs(metric)
    else:
        config = read_json.read_json(config_path)

    config_schema = config_utils.get_config_schema(metric)
    jsonschema.validate(config, config_schema)

    # argument: output
    output_path = None
    if args.output:
        output_path = os.path.join(abs_dir, args.output)

    # argument: detail
    detail_dir = None
    if args.detail:
        detail_dir = os.path.join(abs_dir, args.detail)

    # argument: parallel
    is_parallel = False
    if args.parallel:
        is_parallel = args.parallel

    # argument: optimize
    optimize_config = None
    if args.optimize:
        optimize_config = read_json.read_json(args.optimize)

    return gold_swc_tree, test_swc_trees, test_swc_paths, metric, output_path, detail_dir, config, is_debug, is_parallel, optimize_config


def excute_metric(metric, gold_swc_tree, test_swc_tree, config, detail_dir, output_path, metric_method):
    test_swc_name = test_swc_tree.get_name()

    result, res_gold_swc_tree, res_test_swc_tree = metric_method(
        gold_swc_tree=gold_swc_tree, test_swc_tree=test_swc_tree, config=config
    )
    screen_output = config_utils.get_screen_output()
    result_info = ""
    for key in result:
        if key in screen_output[metric]:
            result_info += "{} = {}\n".format(key.ljust(15, " "), result[key])

    print("---------------Result---------------\n" +
          "swc_file_name   = {}\n".format(test_swc_name) +
          result_info +
          "----------------End-----------------\n"
          )

    base_file_name = test_swc_name[:-4] + "_" + metric + "_"

    def save_detail(swc_tree, file_name):
        ok = swc_save(
            swc_tree=swc_tree,
            out_path=os.path.join(detail_dir, file_name),
            extra=anno_utils.get_detail_type(metric),
        )
        if ok:
            print("[Info:] Details: {} saved".format(file_name))
        else:
            print("[Warning:] Failed to save details: {}".format(file_name))

    if detail_dir:
        if res_gold_swc_tree is not None:
            save_detail(res_gold_swc_tree, base_file_name+"recall.swc")

        if res_test_swc_tree is not None:
            save_detail(res_test_swc_tree, base_file_name+"precision.swc")

    if output_path:
        ok = read_json.save_json(data=result, json_file_path=output_path)
        if ok:
            print("[Info:] Output saved")
        else:
            print("[Warning:] Failed to save output")


# command program
def run():
    abs_dir = os.path.abspath("")
    import_metrics()
    init(abs_dir)

    args = read_parameters()
    gold_swc_tree, test_swc_trees, test_swc_paths, metric, output_path, detail_dir, \
    config, is_debug, is_parallel, optimize_config = set_configs(abs_dir, args)

    metric_manager = get_metric_manager()
    metric_method = metric_manager.get_metric_method(metric)

    if optimize_config is not None:
        optimize.optimize(gold_swc_tree=gold_swc_tree, test_swc_paths=test_swc_paths,
                          optimize_config=optimize_config, metric_config=config, metric_method=metric_method)
    elif is_parallel:
        # use multi process
        max_procs = cpu_count()
        if len(test_swc_trees) < max_procs:
            max_procs = len(test_swc_trees)
        p_pool = Pool(max_procs)
        for test_swc_tree in test_swc_trees:
            p_pool.apply_async(
                excute_metric,
                args=(metric, gold_swc_tree, test_swc_tree, config, detail_dir, output_path, metric_method),
            )
        p_pool.close()
        p_pool.join()
    else:
        for test_swc_tree in test_swc_trees:
            excute_metric(
                metric=metric,
                gold_swc_tree=gold_swc_tree,
                test_swc_tree=test_swc_tree,
                config=config,
                detail_dir=detail_dir,
                output_path=output_path,
                metric_method=metric_method,
            )
    print("Done!")


if __name__ == "__main__":
    sys.exit(run())



import tflib as lib

import numpy as np
import tensorflow as tf

def Layernorm(name, norm_axes, inputs):
    mean, var = tf.nn.moments(inputs, norm_axes, keep_dims=True)

    # Assume the 'neurons' axis is the first of norm_axes. This is the case for fully-connected and BCHW conv layers.
    n_neurons = inputs.get_shape().as_list()[norm_axes[0]]

    offset = lib.param(name+'.offset', np.zeros(n_neurons, dtype='float32'))
    scale = lib.param(name+'.scale', np.ones(n_neurons, dtype='float32'))

    # Add broadcasting dims to offset and scale (e.g. BCHW conv data)
    offset = tf.reshape(offset, [-1] + [1 for i in range(len(norm_axes)-1)])
    scale = tf.reshape(scale, [-1] + [1 for i in range(len(norm_axes)-1)])

    result = tf.nn.batch_normalization(inputs, mean, var, offset, scale, 1e-5)

    return result


# import the necessary packages
import numpy as np
import cv2

def translate(image, x, y):
	# define the translation matrix and perform the translation
	M = np.float32([[1, 0, x], [0, 1, y]])
	shifted = cv2.warpAffine(image, M, (image.shape[1], image.shape[0]))

	# return the translated image
	return shifted

def rotate(image, angle, center=None, scale=1.0):
	# grab the dimensions of the image
	(h, w) = image.shape[:2]

	# if the center is None, initialize it as the center of
	# the image
	if center is None:
		center = (w / 2, h / 2)

	# perform the rotation
	M = cv2.getRotationMatrix2D(center, angle, scale)
	rotated = cv2.warpAffine(image, M, (w, h))

	# return the rotated image
	return rotated

def resize(image, width=None, height=None, inter=cv2.INTER_AREA):
	# initialize the dimensions of the image to be resized and
	# grab the image size
	dim = None
	(h, w) = image.shape[:2]

	# if both the width and height are None, then return the
	# original image
	if width is None and height is None:
		return image

	# check to see if the width is None
	if width is None:
		# calculate the ratio of the height and construct the
		# dimensions
		r = height / float(h)
		dim = (int(w * r), height)

	# otherwise, the height is None
	else:
		# calculate the ratio of the width and construct the
		# dimensions
		r = width / float(w)
		dim = (width, int(h * r))

	# resize the image
	resized = cv2.resize(image, dim, interpolation=inter)

	# return the resized image
	return resized

def skeletonize(image, size, structuring=cv2.MORPH_RECT):
	# determine the area (i.e. total number of pixels in the image),
	# initialize the output skeletonized image, and construct the
	# morphological structuring element
	area = image.shape[0] * image.shape[1]
	skeleton = np.zeros(image.shape, dtype="uint8")
	elem = cv2.getStructuringElement(structuring, size)

	# keep looping until the erosions remove all pixels from the
	# image
	while True:
		# erode and dilate the image using the structuring element
		eroded = cv2.erode(image, elem)
		temp = cv2.dilate(eroded, elem)

		# subtract the temporary image from the original, eroded
		# image, then take the bitwise 'or' between the skeleton
		# and the temporary image
		temp = cv2.subtract(image, temp)
		skeleton = cv2.bitwise_or(skeleton, temp)
		image = eroded.copy()

		# if there are no more 'white' pixels in the image, then
		# break from the loop
		if area == area - cv2.countNonZero(image):
			break

	# return the skeletonized image
	return skeleton

def opencv2matplotlib(image):
	# OpenCV represents images in BGR order; however, Matplotlib
	# expects the image in RGB order, so simply convert from BGR
	# to RGB and return
	return cv2.cvtColor(image, cv2.COLOR_BGR2RGB)


import os
import csv
import unittest
import time
import numpy as np
import multiprocessing as mp

from pyneval.metric.utils.klib import TiffFile
from pyneval.io import read_json
from pyneval.model import swc_node
from pyneval.metric import length_metric as lm
from pyneval.metric import diadem_metric as dm
from pyneval.metric import critical_node_metric as cnm
from pyneval.metric import link_metric as lkm
from pyneval.metric import ssd_metric as ssdm
from pyneval.metric import volume_metric as vm

from test.test_model.visualize.linechart_visual import linechar_visual

import sys

CPU_CORE_NUM = 12


def csv_to_pic_single(csv_path, pic_path, label_list=[]):
    stds, avgs = [], []
    if os.path.isdir(csv_path):
        return None
    fg = 0
    with open(csv_path, 'r', newline='') as f:
        csv_reader = csv.reader(f)
        for row in csv_reader:
            try:
                if len(row) == 0:
                    continue
                if fg == 0:
                    avgs.append(list(map(float, row)))
                else:
                    stds.append(list(map(float, row)))
                fg ^= 1
            except:
                pass

    linechar_visual(
        avgs=avgs,
        stds=stds,
        output_dir=pic_path,
        file_name=os.path.basename(csv_path)[:-4] + '.svg',
        lables=label_list,
    )


def csv_to_pic(csv_path, pic_path, label_list=[]):
    for csv_file in os.listdir(csv_path):
        csv_to_pic_single(os.path.join(csv_path, csv_file), pic_path, label_list=label_list)


def res_to_csv(swc_file, swc_res, csv_path=None, concern_list=[]):
    avgs = []
    stds = []
    for i in range(len(concern_list)):
        avgs.append([])
        stds.append([])

    for raw_per_data in swc_res:
        np_raw_per_data = np.array(raw_per_data)
        avg = np_raw_per_data.mean(axis=0)
        std = np_raw_per_data.std(axis=0)
        for i in range(len(avgs)):
            avgs[i].append(avg[concern_list[i]])
            stds[i].append(std[concern_list[i]])

    if csv_path is not None:
        with open(os.path.join(csv_path, swc_file+".csv"), 'w') as f:
            c_writer = csv.writer(f)
            for i in range(len(avgs)):
                c_writer.writerow(avgs[i])
                c_writer.writerow(stds[i])


# def test_template(func_method, func_args, gold_dir, test_dir, generate_method="volume_move"):
#     '''
#     This function calculate the metric result on all test data of one metric method
#     '''
#     sys.setrecursionlimit(1000000)
#     # res
#     metric_results = {}
#     # choose the modified way: "move" or "delete"
#     # test_method_dir sample: test_dir/move
#     test_method_dir = os.path.join(test_dir, generate_method)
#
#     for swc_data in os.listdir(test_method_dir):
#         # swc_dir sample: test_dir/volume_move/a
#         test_method_swc_dir = os.path.join(test_method_dir, swc_data)
#
#         gold_tree = swc_node.SwcTree()
#         gold_tree.load(os.path.join(gold_dir, swc_data+".swc"))
#
#         swc_res = []
#         for per in os.listdir(test_method_swc_dir):
#             # final_dir sample: test_dir/volume_move/a/10
#             final_dir = os.path.join(test_method_swc_dir, per)
#
#             results = []
#             pool = mp.Pool(processes=CPU_CORE_NUM)
#             for file in os.listdir(final_dir):
#                 test_tree = swc_node.SwcTree()
#                 test_tree.load(os.path.join(final_dir, file))
#                 results.append(
#                     pool.apply_async(func_method, args=tuple([test_tree, gold_tree])+func_args)
#                 )
#                 # results.append(func_method(gold_tree, test_tree,func_args[0], func_args[1]))
#             pool.close()
#             pool.join()
#             clean_result = []
#             for result in results:
#                 clean_result.append(list(result.get()))
#             swc_res.append(clean_result)
#             print("{}----{}".format(swc_data, per))
#         metric_results[swc_data] = swc_res
#     return metric_results


def single_test(test_metric, metric_args, gold_file, test_file_path):
    '''
    This function calculates the metric result between 10% and 100% change when metric method and data is chosen.
    Args:
        test_metric(func): function name of metric to test:
        metric_args(tuple): arguments of tested metric
        file(str): tested file name
        gold_dir(str): path of gold file
        test_file_path(str): basic path of test file
        Example:
            test_file_path:  test_dir/generate_method/file
    Result:
        result(list):
            [
                [average_arg1_10%, average_arg1_20%,..., average_arg1_100%],
                [std_arg1_10%, std_arg1_20%,..., std_arg1_100%],
                [average_arg2_10%, average_arg2_20%,..., average_arg2_100%],
                [std_arg2_10%, std_arg2_20%,..., std_arg2_100%],
                ...
            ]
    '''
    gold_tree = swc_node.SwcTree()
    gold_tree.load(gold_file)

    swc_res = []
    for percentage in os.listdir(test_file_path):
        print("{}---{}".format(gold_file, percentage))
        # final_dir sample: test_dir/volume_move/a/10
        final_dir = os.path.join(test_file_path, percentage)

        per_res = []
        pool = mp.Pool(processes=CPU_CORE_NUM)
        for file in os.listdir(final_dir):
            test_tree = swc_node.SwcTree()
            test_tree.load(os.path.join(final_dir, file))
            per_res.append(
                pool.apply_async(test_metric, args=tuple([gold_tree, test_tree])+metric_args)
            )
            # results.append(func_method(gold_tree, test_tree,func_args[0], func_args[1]))
        pool.close()
        pool.join()
        result_values = []
        for result in per_res:
            result_values.append(list(result.get().values()))
        swc_res.append(result_values)
    return swc_res


def batch_test(test_datas, gold_dir, test_dir, method):
    metric_names = ["ssd_metric", "length_metric", "diadem_metric", "critical_node_metric", "link_metric"]
    metrics = {
        "ssd_metric": ssdm.ssd_metric,
        "length_metric": lm.length_metric,
        "diadem_metric": dm.diadem_metric,
        "critical_node_metric": cnm.critical_node_metric,
        "link_metric": lkm.link_metric}
    configs = {
        "ssd_metric": read_json.read_json("..\\..\\config\\ssd_metric.json"),
        "length_metric": read_json.read_json("..\\..\\config\\length_metric.json"),
        "diadem_metric": read_json.read_json("..\\..\\config\\diadem_metric.json"),
        "critical_node_metric": read_json.read_json("..\\..\\config\\critical_node_metric.json"),
        "link_metric": read_json.read_json("..\\..\\config\\link_metric.json")}
    labels = {
        "ssd_metric": ["ssd_score", "recall", "precision"],
        "length_metric": ["recall", "precision"],
        "diadem_metric": ["final_score"],
        "critical_node_metric": ["true_positive_number", "false_negative_num", "false_positive_num", "matched_mean_distance"],
        "link_metric": ["edge_loss", "tree_dis_loss"]
    }
    metric_result = {}
    test_method_dir = os.path.join(test_dir, method)
    for metric in metric_names:
        for data in test_datas:
            swc_res = single_test(test_metric=metrics[metric],
                                  metric_args=(configs[metric], ),
                                  gold_file=os.path.join(gold_dir, data+".swc"),
                                  test_file_path=os.path.join(test_method_dir, data))
            metric_result[metric+"_"+data+"_"+method] = swc_res
            res_to_csv(swc_file=data,
                       swc_res=swc_res,
                       csv_path=os.path.join(out_dir, "csv_dir"))


if __name__=='__main__':
    sys.setrecursionlimit(10000000)
    metric_names = ["ssd_metric", "length_metric", "diadem_metric", "critical_node_metric", "link_metric"]
    labels = {
        "ssd_metric": ["ssd_score", "recall", "precision"],
        "length_metric": ["recall", "precision"],
        "diadem_metric": ["final_score"],
        "critical_node_metric": ["true_positive_number", "false_negative_num", "false_positive_num", "matched_mean_distance"],
        "link_metric": ["edge_loss", "tree_dis_loss"]
    }

    gold_dir = "../../data/example_selected"
    test_dir = "../../output/random_data"
    out_dir = "../../output/"
    config = read_json.read_json("..\\..\\config\\ssd_metric.json")
    method = "diadem"
    # batch_test()
    # res = single_test(test_metric=ssdm.ssd_metric,
    #                   metric_args=(config, ),
    #                   test_file_path="E:\\00_project\\00_neural_reconstruction\\01_project\PyNeval\output\\random_data\move\\a",
    #                   gold_file="E:\\00_project\\00_neural_reconstruction\\01_project\PyNeval\data\example_selected\\a.swc")
    # res_to_csv("a_ssd", res,
    #            csv_path="E:\\00_project\\00_neural_reconstruction\\01_project\PyNeval\output\csv_dir",
    #            concern_list=[0,1,2])
    for metric in metric_names:
        csv_to_pic(csv_path=os.path.join("E:\\00_project\\00_neural_reconstruction\\01_project\PyNeval\output\csv_dir", metric[:-7]),
                   pic_path="E:\\00_project\\00_neural_reconstruction\\01_project\PyNeval\output\pic_dir",
                   label_list=labels[metric])


from functools import reduce
from Obj3D import Point3D, Sphere, Cone, calculateBound, calScaleRatio
import numpy as np
from numpy import linalg as LA
from scipy.spatial import distance_matrix

def getObjList(nodes, graph, node_idx=None):
    if node_idx:
        # 球体索引列表
        sphere_idxs = [node_idx]+list(graph[node_idx])

        sphere_list = [Sphere(Point3D(*nodes[x].pos), nodes[x].r) for x in sphere_idxs]

        # 椎体索引对列表
        cone_idx_pairs  = [(node_idx, x) for x in graph[node_idx]]

        cone_list = [Cone(Point3D(*nodes[p[0]].pos),nodes[p[0]].r,Point3D(*nodes[p[1]].pos),nodes[p[1]].r) for p in cone_idx_pairs]
    else: # Returen all nodes
        sphere_list=[]
        cone_list=[]
        for node_idx in nodes.keys():
            # 加入当前节点对应的球体
            sphere_list.append(Sphere(Point3D(*nodes[node_idx].pos), nodes[node_idx].r))

            # 椎体索引对列表
            cone_idx_pairs  = [(node_idx, x) for x in graph[node_idx] if node_idx<x]
            cone_list_local = [Cone(Point3D(*nodes[p[0]].pos),nodes[p[0]].r,Point3D(*nodes[p[1]].pos),nodes[p[1]].r) \
                               for p in cone_idx_pairs]
            cone_list.extend(cone_list_local)

    return sphere_list, cone_list

def checkSphereV2(mark, sphere, img_shape):
    bbox = list(sphere.calBBox()) # xmin,ymin,zmin,xmax,ymax,zmax
    for i in range(3):
        j = i+3
        if (bbox[i]<0):
            bbox[i] = 0
        if (bbox[j]>img_shape[i]):
            bbox[j] = img_shape[i]
    (xmin,ymin,zmin,xmax,ymax,zmax) = tuple(bbox)
    (x_idxs,y_idxs,z_idxs)=np.where(mark[xmin:xmax,ymin:ymax,zmin:zmax]==0)
    # points=img_idxs[:3, xmin+x_idxs, ymin+y_idxs, zmin+z_idxs] # 3*M
    # points=points.T # M*3
    xs = np.asarray(xmin+x_idxs).reshape((len(x_idxs),1))
    ys = np.asarray(ymin+y_idxs).reshape((len(y_idxs),1))
    zs = np.asarray(zmin+z_idxs).reshape((len(z_idxs),1))
    points=np.hstack((xs,ys,zs))

    sphere_c_mat = np.array([sphere.center_point.toList()]) # 1*3
    # 计算所有点到所有球心的距离
    dis_mat = distance_matrix(points,sphere_c_mat) # M*1

    # 判断距离是否小于半径
    res_idxs = np.where(dis_mat<=sphere.radius)[0]
    mark[xmin+x_idxs[res_idxs], ymin+y_idxs[res_idxs], zmin+z_idxs[res_idxs]] = 255

def checkConeV2(mark, cone, img_shape):
    bbox = list(cone.calBBox()) # xmin,ymin,zmin,xmax,ymax,zmax
    for i in range(3):
        j = i+3
        if (bbox[i]<0):
            bbox[i] = 0
        if (bbox[j]>img_shape[i]):
            bbox[j] = img_shape[i]
    (xmin,ymin,zmin,xmax,ymax,zmax) = tuple(bbox)

    (x_idxs,y_idxs,z_idxs)=np.where(mark[xmin:xmax,ymin:ymax,zmin:zmax]==0)
    # points=img_idxs[:, xmin+x_idxs, ymin+y_idxs, zmin+z_idxs] # 3*M
    # points=points.T # M*3
    xs = np.asarray(xmin+x_idxs).reshape((len(x_idxs),1))
    ys = np.asarray(ymin+y_idxs).reshape((len(y_idxs),1))
    zs = np.asarray(zmin+z_idxs).reshape((len(z_idxs),1))
    ns = np.ones((len(z_idxs),1))
    points=np.hstack((xs,ys,zs,ns))

    # 每个圆锥的还原矩阵
    r_min=cone.up_radius
    r_max=cone.bottom_radius
    height=cone.height
    cone_revert_mat = cone.revertMat().T # 4*4

    # 每个椎体还原后坐标
    revert_coor_mat = np.matmul(points, cone_revert_mat) # M*4
    revert_radius_list = LA.norm(revert_coor_mat[:,:2], axis=1) # M

    # Local Indexs
    M = points.shape[0]
    l_idx = np.arange(M) # M (1-dim)
    l_mark = np.ones((M,), dtype=bool)

    # 过滤高度在外部的点
    res_idxs = np.logical_or(revert_coor_mat[l_idx[l_mark],2]<0, revert_coor_mat[l_idx[l_mark],2]>height)
    l_mark[l_idx[l_mark][res_idxs]]=False

    # 过滤半径在外部的点
    res_idxs = revert_radius_list[l_idx[l_mark]]>r_max
    l_mark[l_idx[l_mark][res_idxs]]=False

    # 过滤半径在内部的点
    res_idxs = revert_radius_list[l_idx[l_mark]]<=r_min
    mark[xmin+x_idxs[l_idx[l_mark][res_idxs]], ymin+y_idxs[l_idx[l_mark][res_idxs]], zmin+z_idxs[l_idx[l_mark][res_idxs]]] = 255
    l_mark[l_idx[l_mark][res_idxs]]=False

    # 计算剩余
    # import pdb
    # pdb.set_trace();
    if r_max>r_min:
        res_idxs = ((r_max-revert_radius_list[l_idx[l_mark]])*height/(r_max-r_min)) >= revert_coor_mat[l_idx[l_mark],2]
        mark[xmin+x_idxs[l_idx[l_mark][res_idxs]], ymin+y_idxs[l_idx[l_mark][res_idxs]], zmin+z_idxs[l_idx[l_mark][res_idxs]]] = 255
        l_mark[l_idx[l_mark][res_idxs]]=False


#@profile
def draw3dStackSparseV2(sphere_list, cone_list, img_shape):
    '''draw3dStack: Draw 3D image stack.
    Args:
        param1 (int): The first parameter.

    Returns:
        bool: The return value.
    '''
    # print('img_shape', img_shape)
    img_total_length = reduce(lambda x,y:x*y, img_shape)

    # 创建原始矩阵
    mark = np.zeros(img_shape, dtype=np.uint8)

    # 对球体进行判断
    for s in sphere_list:
        checkSphereV2(mark, s, img_shape);
    # 对圆台进行判断
    for c in cone_list:
        checkConeV2(mark, c, img_shape);
    ## 绘制
    #mark=np.where(mark==1, 255, 0).astype(np.uint8)
    mark=np.swapaxes(mark,0,2)
    return mark



import os


def is_path_valid(file_path):
    tmp_path, tmp_file = os.path.split(file_path)

    if not os.path.exists(file_path):
        if not os.path.exists(tmp_path):
            os.makedirs(tmp_path)
        new_file = open(file_path, "w")
        new_file.close()

    # if os.path.isdir(file_path):
    # raise PyNevalError("[Error: ] file: \"{}\" has exist in path: \"{}\". and it is a menu"
    #                 .format(tmp_file, tmp_path))
    # return False

    return not os.path.isdir(file_path)


def save_line_tuple_as_swc(match_fail, file_path):
    with open(file_path, "w") as f:
        f.truncate()
        f.write("# total unmatched edges: {}\n".format(len(match_fail)))
        for line_tuple in match_fail:
            f.write(line_tuple[0].to_swc_str())
        f.write("# --End--\n")


def save_as_swc(object, file_path):
    if not is_path_valid(file_path):
        return False

    if isinstance(object, set):
        save_line_tuple_as_swc(object, file_path)


def print_line_tuple_swc(match_fail_set):
    for line_tuple in match_fail_set:
        print ("pos_1: {}, pos_2 {}".format(line_tuple[0]._pos, line_tuple[1]._pos))


def print_swc(object):
    if isinstance(object, set):
        print_line_tuple_swc(object)


def swc_save(swc_tree, out_path, extra=None):
    out_path = os.path.normpath(out_path)
    out_dir = os.path.dirname(out_path)

    while not os.path.exists(out_dir):
        print('[Info]: "{}" dose not exist. Create new path?[y/n]'.format(out_dir))
        choice = input()
        if choice.lower() == 'y':
            os.makedirs(out_dir)
        elif choice.lower() == 'n':
            print('[Info]: "{}" dose not exist. Input new directory?[y/n]'.format(out_dir))
            choice = input()
            if choice.lower() == 'y':
                print('[Info]: Input new directory')
                out_dir = input()
            elif choice.lower() == 'n':
                return False
            else:
                continue
    out_new_path = os.path.join(out_dir, os.path.basename(out_path))
    while not out_new_path[-4:] == '.swc':
        print('[Info]: "{}"  is not a swc file. Input another path?[y/n]'.format(out_new_path))
        choice = input()
        if choice.lower() == 'y':
            print('[Info]: Input new path:')
            out_new_path = input()
        elif choice.lower() == 'n':
            return False

    while os.path.exists(out_new_path):
        print('[Info]: "{}" is already existed. Overwrite?[y/n]'.format(out_new_path))
        choice = input()
        if choice.lower() == 'y':
            break
        elif choice.lower() == 'n':
            return False

    swc_node_list = swc_tree.get_node_list()
    swc_tree.sort_node_list(key="id")
    with open(out_new_path, "w") as f:
        f.truncate()
        if extra is not None:
            f.write(extra)
        for node in swc_node_list:
            if node.is_virtual():
                continue
            try:
                f.write(
                    "{} {} {} {} {} {} {}\n".format(
                        node.get_id(),
                        node._type,
                        node.get_x(),
                        node.get_y(),
                        node.get_z(),
                        node.radius(),
                        node.parent.get_id(),
                    )
                )
            except:
                continue
    return True


import sys
import jsonschema

from pyneval.model import swc_node
from pyneval.metric.utils import edge_match_utils
from pyneval.io import read_swc
from pyneval.io import swc_writer
from pyneval.metric.utils.metric_manager import get_metric_manager
from pyneval.metric.utils import basic_utils

metric_manager = get_metric_manager()


class LengthMetric(object):
    """
    length metric
    """

    def __init__(self, config):
        # read config
        self.radius_mode = config["radius_mode"]
        self.radius_threshold = config["radius_threshold"]
        self.length_threshold = config["length_threshold"]
        self.scale = config["scale"]
        self.debug = config["debug"]

    def length_metric_run(self, gold_swc_tree=None, test_swc_tree=None,
                          radius_threshold=-1.0, length_threshold=0.2):
        """
        get matched edge set and calculate recall and precision
        Args:
            gold_swc_tree(SwcTree):
            test_swc_tree(SwcTree):
            radius_threshold(float): threshold of key point radius
            length_threshold(float): threshold of length of the matching edges
        Returns:
            tuple: contain two values to demonstrate metric result
                precision(float): percentage of total length of edges that are matched compared to test tree
                recall(float): percentage of total length of edges that are matched compared to gold tree
        Raises:
            None
        """
        # get matched edge set
        match_edges, test_match_length = edge_match_utils.get_match_edges(gold_swc_tree=gold_swc_tree,
                                                                          test_swc_tree=test_swc_tree,
                                                                          radius_threshold=radius_threshold,
                                                                          length_threshold=length_threshold,
                                                                          debug=self.debug)
        # calculate the sum of matched length and total length of gold and test tree
        match_length = 0.0
        for line_tuple in match_edges:
            match_length += line_tuple[0].parent_distance()

        gold_total_length = round(gold_swc_tree.length(), 8)
        test_total_length = round(test_swc_tree.length(), 8)
        match_length = round(match_length, 8)
        test_match_length = round(test_match_length, 8)

        if self.debug:
            print("match_length = {}, test_match_length = {}, gold_total_length = {}, test_total_length = {}"
                  .format(match_length, test_match_length, gold_total_length, test_total_length))
        # calculate recall and precision
        if gold_total_length != 0:
            recall = round(match_length / gold_total_length, 8)
        else:
            recall = 0

        if test_total_length != 0:
            precision = round(test_match_length / test_total_length, 8)
        else:
            precision = 0

        return min(recall, 1.0), min(precision, 1.0)

    def run(self, gold_swc_tree, test_swc_tree):
        """Main function of length metric.
            unpack config and run the matching function
            Args:
                gold_swc_tree(SwcTree):
                test_swc_tree(SwcTree):
            Example:
                test_tree = swc_node.SwcTree()
                gold_tree = swc_node.SwcTree()
                gold_tree.load("..\\..\\data\\test_data\\geo_metric_data\\gold_fake_data1.swc")
                test_tree.load("..\\..\\data\\test_data\\geo_metric_data\\test_fake_data1.swc")
                lm_res = length_metric(gold_swc_tree=gold_tree,
                                       test_swc_tree=test_tree,
                                       config=config)
            Returns:
                tuple: contain two values to demonstrate metric result
                    precision(float): percentage of total length of edges that are matched compared to test tree
                    recall(float): percentage of total length of edges that are matched compared to gold tree
            Raises:
                None
            """
        gold_swc_tree.rescale(self.scale)
        test_swc_tree.rescale(self.scale)
        gold_swc_tree.set_node_type_by_topo(root_id=1)
        test_swc_tree.set_node_type_by_topo(root_id=5)

        if self.radius_mode == 1:
            self.radius_threshold *= -1
        # check every edge in test, if it is overlap with any edge in gold three
        recall, precision = self.length_metric_run(gold_swc_tree=gold_swc_tree,
                                                   test_swc_tree=test_swc_tree,
                                                   radius_threshold=self.radius_threshold,
                                                   length_threshold=self.length_threshold,
                                                   )
        if self.debug:
            print("Recall = {}, Precision = {}".format(recall, precision))

        res = {
            "recall": recall,
            "precision": precision,
            "f1_score": basic_utils.get_f1score(recall=recall, precision=precision)
        }
        return res, gold_swc_tree, test_swc_tree


# @do_cprofile("./mkm_run.prof")
@metric_manager.register(
    name="length",
    config="length_metric.json",
    desc="length of matched branches and fibers",
    public=True,
    alias=['ML']
)
def length_metric(gold_swc_tree, test_swc_tree, config):
    """Main function of length metric.
    unpack config and run the matching function
    Args:
        gold_swc_tree(SwcTree):
        test_swc_tree(SwcTree):
        config(Dict):
            keys: the name of configs
            items: config values
    Example:
        test_tree = swc_node.SwcTree()
        gold_tree = swc_node.SwcTree()
        gold_tree.load("..\\..\\data\\test_data\\geo_metric_data\\gold_fake_data1.swc")
        test_tree.load("..\\..\\data\\test_data\\geo_metric_data\\test_fake_data1.swc")
        lm_res = length_metric(gold_swc_tree=gold_tree,
                               test_swc_tree=test_tree,
                               config=config)
    Returns:
        tuple: contain two values to demonstrate metric result
            precision(float): percentage of total length of edges that are matched compared to test tree
            recall(float): percentage of total length of edges that are matched compared to gold tree
    Raises:
        None
    """

    length_metric = LengthMetric(config)
    return length_metric.run(gold_swc_tree, test_swc_tree)
